# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Definition of known configuration options and methods to parse and get option values."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import collections
import six

__all__ = ('get_option', 'get_option_names', 'parse_option')


class NO_DEFAULT_PLACEHOLDER(object):  # pylint: disable=too-few-public-methods,invalid-name,useless-object-inheritance
    """A dummy class to serve as a marker for no default being specified in the `get_option` function."""


NO_DEFAULT = NO_DEFAULT_PLACEHOLDER()
DEFAULT_DAEMON_TIMEOUT = 20  # Default timeout in seconds for circus client calls
VALID_LOG_LEVELS = ['CRITICAL', 'ERROR', 'WARNING', 'REPORT', 'INFO', 'DEBUG']

Option = collections.namedtuple('Option', ['name', 'key', 'valid_type', 'valid_values', 'default', 'description'])


def get_option(option_name):
    """Return a configuration option.configuration

    :param option_name: the name of the configuration option
    :return: the configuration option
    :raises ValueError: if the configuration option does not exist
    """
    try:
        option = Option(option_name, **CONFIG_OPTIONS[option_name])
    except KeyError:
        raise ValueError('the option {} does not exist'.format(option_name))
    else:
        return option


def get_option_names():
    """Return a list of available option names.

    :return: list of available option names
    """
    return CONFIG_OPTIONS.keys()


def parse_option(option_name, option_value):
    """Parse and validate a value for a configuration option.

    :param option_name: the name of the configuration option
    :param option_value: the option value
    :return: a tuple of the option and the parsed value
    """
    option = get_option(option_name)

    value = False

    if option.valid_type == 'bool':
        if isinstance(option_value, six.string_types):
            if option_value.strip().lower() in ['0', 'false', 'f']:
                value = False
            elif option_value.strip().lower() in ['1', 'true', 't']:
                value = True
            else:
                raise ValueError('option {} expects a boolean value'.format(option.name))
        else:
            value = bool(option_value)
    elif option.valid_type == 'string':
        value = six.text_type(option_value)
    elif option.valid_type == 'int':
        value = int(option_value)
    elif option.valid_type == 'list_of_str':
        value = option_value.split()
    else:
        raise NotImplementedError('Type string {} not implemented yet'.format(option.valid_type))

    if option.valid_values is not None:
        if value not in option.valid_values:
            raise ValueError('{} is not among the list of accepted values for option {}'.format(value, option.name))

    return option, value


CONFIG_OPTIONS = {
    'runner.poll.interval': {
        'key': 'runner_poll_interval',
        'valid_type': 'int',
        'valid_values': None,
        'default': 1,
        'description': 'The polling interval in seconds to be used by process runners',
    },
    'daemon.timeout': {
        'key': 'daemon_timeout',
        'valid_type': 'int',
        'valid_values': None,
        'default': DEFAULT_DAEMON_TIMEOUT,
        'description': 'The timeout in seconds for calls to the circus client',
    },
    'verdi.shell.auto_import': {
        'key': 'verdi_shell_auto_import',
        'valid_type': 'string',
        'valid_values': None,
        'default': '',
        'description': 'Additional modules/functions/classes to be automatically loaded in `verdi shell`',
    },
    'logging.aiida_loglevel': {
        'key': 'logging_aiida_log_level',
        'valid_type': 'string',
        'valid_values': VALID_LOG_LEVELS,
        'default': 'REPORT',
        'description': 'Minimum level to log to daemon log and the `DbLog` table for the `aiida` logger',
    },
    'logging.db_loglevel': {
        'key': 'logging_db_log_level',
        'valid_type': 'string',
        'valid_values': VALID_LOG_LEVELS,
        'default': 'REPORT',
        'description': 'Minimum level to log to the DbLog table',
    },
    'logging.tornado_loglevel': {
        'key': 'logging_tornado_log_level',
        'valid_type': 'string',
        'valid_values': VALID_LOG_LEVELS,
        'default': 'WARNING',
        'description': 'Minimum level to log to daemon log and the `DbLog` table for the `tornado` logger',
    },
    'logging.plumpy_loglevel': {
        'key': 'logging_plumpy_log_level',
        'valid_type': 'string',
        'valid_values': VALID_LOG_LEVELS,
        'default': 'WARNING',
        'description': 'Minimum level to log to daemon log and the `DbLog` table for the `plumpy` logger',
    },
    'logging.kiwipy_loglevel': {
        'key': 'logging_kiwipy_log_level',
        'valid_type': 'string',
        'valid_values': VALID_LOG_LEVELS,
        'default': 'WARNING',
        'description': 'Minimum level to log to daemon log and the `DbLog` table for the `kiwipy` logger',
    },
    'logging.paramiko_loglevel': {
        'key': 'logging_paramiko_log_level',
        'valid_type': 'string',
        'valid_values': VALID_LOG_LEVELS,
        'default': 'WARNING',
        'description': 'Minimum level to log to daemon log and the `DbLog` table for the `paramiko` logger',
    },
    'logging.alembic_loglevel': {
        'key': 'logging_alembic_log_level',
        'valid_type': 'string',
        'valid_values': VALID_LOG_LEVELS,
        'default': 'WARNING',
        'description': 'Minimum level to log to daemon log and the `DbLog` table for the `alembic` logger',
    },
    'logging.sqlalchemy_loglevel': {
        'key': 'logging_sqlalchemy_loglevel',
        'valid_type': 'string',
        'valid_values': VALID_LOG_LEVELS,
        'default': 'WARNING',
        'description': 'Minimum level to log to daemon log and the `DbLog` table for the `sqlalchemy` logger',
    },
    'logging.circus_loglevel': {
        'key': 'logging_circus_log_level',
        'valid_type': 'string',
        'valid_values': VALID_LOG_LEVELS,
        'default': 'INFO',
        'description': 'Minimum level to log to daemon log and the `DbLog` table for the `circus` logger',
    },
    'tcod.depositor_username': {
        'key': 'tcod_depositor_username',
        'valid_type': 'string',
        'valid_values': None,
        'default': NO_DEFAULT,
        'description': 'Username for TCOD deposition',
    },
    'tcod.depositor_password': {
        'key': 'tcod_depositor_password',
        'valid_type': 'string',
        'valid_values': None,
        'default': NO_DEFAULT,
        'description': 'Password for TCOD deposition',
    },
    'tcod.depositor_email': {
        'key': 'tcod_depositor_email',
        'valid_type': 'string',
        'valid_values': None,
        'default': NO_DEFAULT,
        'description': 'E-mail address for TCOD deposition',
    },
    'tcod.depositor_author_name': {
        'key': 'tcod_depositor_author_name',
        'valid_type': 'string',
        'valid_values': None,
        'default': NO_DEFAULT,
        'description': 'Author name for TCOD depositions',
    },
    'tcod.depositor_author_email': {
        'key': 'tcod_depositor_author_email',
        'valid_type': 'string',
        'valid_values': None,
        'default': NO_DEFAULT,
        'description': 'E-mail address for TCOD depositions',
    },
    'warnings.showdeprecations': {
        'key': 'show_deprecations',
        'valid_type': 'bool',
        'valid_values': None,
        'default': True,
        'description': 'Boolean whether to print AiiDA deprecation warnings',
    },
}



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the Archive class."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os

from aiida.backends.testbase import AiidaTestCase
from aiida.common.archive import Archive
from aiida.common.exceptions import InvalidOperation


def get_archive_file(archive):
    """
    Return the absolute path of the archive file used for testing purposes. The expected path for these files:

        aiida.backends.tests.export_import_test_files.migrate

    :param archive: the relative filename of the archive
    :returns: absolute filepath of the archive test file
    """
    dirpath_current = os.path.dirname(os.path.abspath(__file__))
    dirpath_archive = os.path.join(dirpath_current, os.pardir, 'fixtures', 'export', 'migrate')

    return os.path.join(dirpath_archive, archive)


class TestCommonArchive(AiidaTestCase):
    """Tests for the :class:`aiida.common.archive.Archive` class."""

    def test_context_required(self):
        """Verify that accessing a property of an Archive outside of a context manager raises."""
        with self.assertRaises(InvalidOperation):
            filepath = get_archive_file('export_v0.1.aiida')
            archive = Archive(filepath)
            _ = archive.version_format

    def test_version_format(self):
        """Verify that `version_format` return the correct archive format version."""
        filepath = get_archive_file('export_v0.1.aiida')
        with Archive(filepath) as archive:
            self.assertEqual(archive.version_format, '0.1')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""AiiDA manager for global settings"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import functools

from .configuration import get_config

__all__ = ('get_manager', 'reset_manager')

MANAGER = None


class Manager(object):  # pylint: disable=useless-object-inheritance
    """
    Manager singleton to provide global versions of commonly used profile/settings related objects
    and methods to facilitate their construction.

    In AiiDA the settings of many objects are tied to options defined in the current profile.  This
    means that certain objects should be constructed in a way that depends on the profile.  Instead of
    having disparate parts of the code accessing the profile we put together here the profile and methods
    to create objects based on the current settings.

    It is also a useful place to put objects where there can be a single 'global' (per profile) instance.

    Future plans:
      * reset manager cache when loading a new profile
    """

    def _load_backend(self, schema_check=True):
        """Load the backend for the currently configured profile and return it.

        :param schema_check: force a database schema check if the database environment has not yet been loaded
        :return: the database backend
        :rtype: :class:`aiida.orm.Backend`
        """
        from aiida.backends.profile import BACKEND_DJANGO, BACKEND_SQLA
        from aiida.backends.utils import is_dbenv_loaded, load_dbenv, _load_dbenv_noschemacheck

        profile = self.get_profile()

        if not is_dbenv_loaded():
            if schema_check:
                load_dbenv(profile.name)
            else:
                _load_dbenv_noschemacheck(profile.name)

        backend_type = profile.dictionary.AIIDADB_BACKEND

        if backend_type == BACKEND_DJANGO:
            from aiida.orm.implementation.django.backend import DjangoBackend
            self._backend = DjangoBackend()
        elif backend_type == BACKEND_SQLA:
            from aiida.orm.implementation.sqlalchemy.backend import SqlaBackend
            self._backend = SqlaBackend()
        else:
            raise RuntimeError('Invalid backend type {} in profile: {}'.format(backend_type, profile.name))

        return self._backend

    def get_backend(self):
        """
        Get the database backend

        :return: the database backend
        :rtype: :class:`aiida.orm.Backend`
        """
        if self._backend is None:
            self._load_backend()

        return self._backend

    def get_profile(self):
        """
        Get the current profile

        :return: current loaded profile instance
        :rtype: :class:`~aiida.manage.configuration.profile.Profile`
        """
        if self._profile is None:
            config = get_config()
            self._profile = config.current_profile

        return self._profile

    def get_persister(self):
        """
        Get the persister

        :return: the current persister instance
        :rtype: :class:`plumpy.Persister`
        """
        from aiida.engine import persistence

        if self._persister is None:
            self._persister = persistence.AiiDAPersister()

        return self._persister

    def get_communicator(self):
        """
        Get the communicator

        :return: a global communicator instance
        :rtype: :class:`kiwipy.Communicator`
        """
        if self._communicator is None:
            self._communicator = self.create_communicator()

        return self._communicator

    def create_communicator(self, task_prefetch_count=None, with_orm=True):
        """
        Create a Communicator

        :param task_prefetch_count: optional specify how many tasks this communicator take simultaneously
        :param with_orm: if True, use ORM (de)serializers. If false, use json.
            This is used by verdi status to get a communicator without needing to load the dbenv.

        :return: the communicator instance
        :rtype: :class:`~kiwipy.rmq.communicator.RmqThreadCommunicator`
        """
        from aiida.manage.external import rmq
        import kiwipy.rmq
        profile = self.get_profile()

        if task_prefetch_count is None:
            task_prefetch_count = rmq._RMQ_TASK_PREFETCH_COUNT  # pylint: disable=protected-access

        url = rmq.get_rmq_url()
        prefix = profile.rmq_prefix

        # This needs to be here, because the verdi commands will call this function and when called in unit tests the
        # testing_mode cannot be set.
        testing_mode = profile.is_test_profile

        message_exchange = rmq.get_message_exchange_name(prefix)
        task_exchange = rmq.get_task_exchange_name(prefix)
        task_queue = rmq.get_launch_queue_name(prefix)

        if with_orm:
            from aiida.orm.utils import serialize
            encoder = functools.partial(serialize.serialize, encoding='utf-8')
            decoder = serialize.deserialize
        else:
            # used by verdi status to get a communicator without needing to load the dbenv
            from aiida.common import json
            encoder = functools.partial(json.dumps, encoding='utf-8')
            decoder = json.loads

        return kiwipy.rmq.RmqThreadCommunicator.connect(
            connection_params={'url': url},
            message_exchange=message_exchange,
            encoder=encoder,
            decoder=decoder,
            task_exchange=task_exchange,
            task_queue=task_queue,
            task_prefetch_count=task_prefetch_count,
            testing_mode=testing_mode,
        )

    def get_daemon_client(self):
        """
        Return the daemon client for the current profile.

        :return: the daemon client
        :rtype: :class:`aiida.daemon.client.DaemonClient`
        :raises aiida.common.MissingConfigurationError: if the configuration file cannot be found
        :raises aiida.common.ProfileConfigurationError: if the given profile does not exist
        """
        from aiida.engine.daemon.client import DaemonClient

        if self._daemon_client is None:
            self._daemon_client = DaemonClient(self.get_profile())

        return self._daemon_client

    def get_process_controller(self):
        """
        Get a process controller

        :return: the process controller instance
        :rtype: :class:`plumpy.RemoteProcessThreadController`
        """
        import plumpy
        if self._process_controller is None:
            self._process_controller = plumpy.RemoteProcessThreadController(self.get_communicator())

        return self._process_controller

    def get_runner(self):
        """
        Get a runner that is based on the current profile settings and can be used globally by the code.

        :return: the global runner
        :rtype: :class:`aiida.engine.runners.Runner`
        """
        if self._runner is None:
            self._runner = self.create_runner()

        return self._runner

    def set_runner(self, new_runner):
        """
        Set the currently used runner

        :param new_runner: the new runner to use
        :type new_runner: :class:`aiida.engine.runners.Runner`
        """
        if self._runner is not None:
            self._runner.close()

        self._runner = new_runner

    def create_runner(self, with_persistence=True, **kwargs):
        """
        Create a new runner

        :param with_persistence: create a runner with persistence enabled
        :type with_persistence: bool
        :return: a new runner instance
        :rtype: :class:`aiida.engine.runners.Runner`
        """
        from aiida.engine import runners

        config = get_config()
        profile = self.get_profile()
        poll_interval = 0.0 if profile.is_test_profile else config.option_get('runner.poll.interval')

        settings = {'rmq_submit': False, 'poll_interval': poll_interval}
        settings.update(kwargs)

        if 'communicator' not in settings:
            # Only call get_communicator if we have to as it will lazily create
            settings['communicator'] = self.get_communicator()

        if with_persistence and 'persister' not in settings:
            settings['persister'] = self.get_persister()

        return runners.Runner(**settings)

    def create_daemon_runner(self, loop=None):
        """
        Create a new daemon runner.  This is used by workers when the daemon is running and in testing.

        :param loop: the (optional) tornado event loop to use
        :type loop: :class:`tornado.ioloop.IOLoop`
        :return: a runner configured to work in the daemon configuration
        :rtype: :class:`aiida.engine.runners.Runner`
        """
        import plumpy
        from aiida.engine import persistence
        from aiida.manage.external import rmq
        runner = self.create_runner(rmq_submit=True, loop=loop)
        runner_loop = runner.loop

        # Listen for incoming launch requests
        task_receiver = rmq.ProcessLauncher(
            loop=runner_loop,
            persister=self.get_persister(),
            load_context=plumpy.LoadSaveContext(runner=runner),
            loader=persistence.get_object_loader())

        def callback(*args, **kwargs):
            return plumpy.create_task(functools.partial(task_receiver, *args, **kwargs), loop=runner_loop)

        runner.communicator.add_task_subscriber(callback)

        return runner

    def close(self):
        """
        Reset the global settings entirely and release any global objects
        """
        if self._communicator is not None:
            self._communicator.stop()
        if self._runner is not None:
            self._runner.stop()

        self._backend = None
        self._config = None
        self._profile = None
        self._communicator = None
        self._daemon_client = None
        self._process_controller = None
        self._persister = None
        self._runner = None

    def __init__(self):
        super(Manager, self).__init__()
        self._backend = None  # type: aiida.orm.Backend
        self._config = None  # type: aiida.manage.configuration.config.Config
        self._daemon_client = None  # type: aiida.daemon.client.DaemonClient
        self._profile = None  # type: aiida.manage.configuration.profile.Profile
        self._communicator = None  # type: kiwipy.rmq.RmqThreadCommunicator
        self._process_controller = None  # type: plumpy.RemoteProcessThreadController
        self._persister = None  # type: aiida.engine.persistence.AiiDAPersister
        self._runner = None  # type: aiida.engine.runners.Runner


def get_manager():
    global MANAGER  # pylint: disable=global-statement
    if MANAGER is None:
        MANAGER = Manager()
    return MANAGER


def reset_manager():
    global MANAGER  # pylint: disable=global-statement
    if MANAGER is not None:
        MANAGER.close()
        MANAGER = None



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the ORM mixin classes."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.testbase import AiidaTestCase
from aiida.common import exceptions
from aiida.common.links import LinkType
from aiida.orm import Int, CalculationNode
from aiida.orm.utils.mixins import Sealable


class TestSealable(AiidaTestCase):
    """Tests for the `Sealable` mixin class."""

    @staticmethod
    def test_change_updatable_attrs_after_store():
        """Verify that a Sealable node can alter updatable attributes even after storing."""

        node = CalculationNode().store()

        for attr in CalculationNode._updatable_attributes:  # pylint: disable=protected-access,not-an-iterable
            if attr != Sealable.SEALED_KEY:
                node.set_attribute(attr, 'a')

    def test_validate_incoming_sealed(self):
        """Verify that trying to add a link to a sealed node will raise."""
        data = Int(1).store()
        node = CalculationNode().store()
        node.seal()

        with self.assertRaises(exceptions.ModificationNotAllowed):
            node.validate_incoming(data, link_type=LinkType.INPUT_CALC, link_label='input')

    def test_validate_outgoing_sealed(self):
        """Verify that trying to add a link from a sealed node will raise."""
        data = Int(1).store()
        node = CalculationNode().store()
        node.seal()

        with self.assertRaises(exceptions.ModificationNotAllowed):
            node.validate_outgoing(data, link_type=LinkType.CREATE, link_label='create')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`Data` sub class to represent a string value."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import six

from .base import BaseType, to_aiida_type

__all__ = ('Str',)


class Str(BaseType):
    """`Data` sub class to represent a string value."""

    _type = str


@to_aiida_type.register(six.string_types[0])
def _(value):
    return Str(value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from six.moves import range

import numpy

from django.apps import apps
from django.db.migrations.executor import MigrationExecutor
from django.db import connection

from aiida.backends.testbase import AiidaTestCase
from aiida.backends.djsite.db.migrations import ModelModifierV0025
from aiida.backends.general.migrations import utils
from aiida.common.exceptions import IntegrityError
from aiida.manage.database.integrity.duplicate_uuid import deduplicate_uuids, verify_uuid_uniqueness


class TestMigrations(AiidaTestCase):
    @property
    def app(self):
        return apps.get_containing_app_config(type(self).__module__).name.split('.')[-1]

    migrate_from = None
    migrate_to = None

    def setUp(self):
        """Go to a specific schema version before running tests."""
        from aiida.orm import autogroup

        self.current_autogroup = autogroup.current_autogroup
        autogroup.current_autogroup = None
        assert self.migrate_from and self.migrate_to, \
            "TestCase '{}' must define migrate_from and migrate_to properties".format(type(self).__name__)
        self.migrate_from = [(self.app, self.migrate_from)]
        self.migrate_to = [(self.app, self.migrate_to)]
        executor = MigrationExecutor(connection)
        self.apps = executor.loader.project_state(self.migrate_from).apps

        # Reverse to the original migration
        executor.migrate(self.migrate_from)

        self.default_user = self.backend.users.create('{}@aiida.net'.format(self.id())).store()
        self.DbNode = self.apps.get_model('db', 'DbNode')

        try:
            self.setUpBeforeMigration()
            # Run the migration to test
            executor = MigrationExecutor(connection)
            executor.loader.build_graph()
            executor.migrate(self.migrate_to)

            self.apps = executor.loader.project_state(self.migrate_to).apps
        except Exception as exception:
            # Bring back the DB to the correct state if this setup part fails
            import traceback
            traceback.print_stack()
            print('EXCEPTION', exception)
            self._revert_database_schema()
            raise

    def tearDown(self):
        """At the end make sure we go back to the latest schema version."""
        from aiida.orm import autogroup
        self._revert_database_schema()
        autogroup.current_autogroup = self.current_autogroup

    def setUpBeforeMigration(self):
        """Anything to do before running the migrations, which should be implemented in test subclasses."""

    def _revert_database_schema(self):
        """Bring back the DB to the correct state."""
        from ..migrations import LATEST_MIGRATION
        self.migrate_to = [(self.app, LATEST_MIGRATION)]
        executor = MigrationExecutor(connection)
        executor.migrate(self.migrate_to)

    def load_node(self, pk):
        return self.DbNode.objects.get(pk=pk)


class TestMigrationsModelModifierV0025(TestMigrations):
    """Sub class of `TestMigrations` that need to work on node attributes using the `ModelModifierV0025`."""

    def set_attribute(self, node, key, value):
        DbAttribute = self.apps.get_model('db', 'DbAttribute')
        modifier = ModelModifierV0025(self.apps, DbAttribute)
        modifier.set_value_for_node(node.pk, key, value)

    def get_attribute(self, node, key, default=None):
        DbAttribute = self.apps.get_model('db', 'DbAttribute')
        modifier = ModelModifierV0025(self.apps, DbAttribute)
        try:
            return modifier.get_value_for_node(node.pk, key)
        except AttributeError:
            return default

    def get_node_array(self, node, name):
        return utils.load_numpy_array_from_repository(node.uuid, name)

    def set_node_array(self, node, name, array):
        """Store a new numpy array inside a node. Possibly overwrite the array if it already existed.

        Internally, it stores a name.npy file in numpy format.

        :param name: The name of the array.
        :param array: The numpy array to store.
        """
        utils.store_numpy_array_in_repository(node.uuid, name, array)
        self.set_attribute(node, 'array|{}'.format(name), list(array.shape))


class TestNoMigrations(AiidaTestCase):

    def test_no_remaining_migrations(self):
        """Verify that no django migrations remain.

        Equivalent to python manage.py makemigrations --check"""

        from django.core.management import call_command

        # Raises SystemExit, if migrations remain
        call_command('makemigrations', '--check', verbosity=0)


class TestDuplicateNodeUuidMigration(TestMigrations):

    migrate_from = '0013_django_1_8'
    migrate_to = '0014_add_node_uuid_unique_constraint'

    def setUpBeforeMigration(self):
        from aiida.common.utils import get_new_uuid
        self.file_name = 'test.temp'
        self.file_content = u'#!/bin/bash\n\necho test run\n'

        self.nodes_boolean = []
        self.nodes_integer = []
        self.n_bool_duplicates = 2
        self.n_int_duplicates = 4

        node_bool = self.DbNode(type='data.bool.Bool.', user_id=self.default_user.id, uuid=get_new_uuid())
        node_bool.save()

        node_int = self.DbNode(type='data.int.Int.', user_id=self.default_user.id, uuid=get_new_uuid())
        node_int.save()

        self.nodes_boolean.append(node_bool)
        self.nodes_integer.append(node_int)

        for i in range(self.n_bool_duplicates):
            node = self.DbNode(type='data.bool.Bool.', user_id=self.default_user.id, uuid=node_bool.uuid)
            node.save()
            utils.put_object_from_string(node.uuid, self.file_name, self.file_content)
            self.nodes_boolean.append(node)

        for i in range(self.n_int_duplicates):
            node = self.DbNode(type='data.int.Int.', user_id=self.default_user.id, uuid=node_int.uuid)
            node.save()
            utils.put_object_from_string(node.uuid, self.file_name, self.file_content)
            self.nodes_integer.append(node)

        # Verify that there are duplicate UUIDs by checking that the following function raises
        with self.assertRaises(IntegrityError):
            verify_uuid_uniqueness(table='db_dbnode')

        # Now run the function responsible for solving duplicate UUIDs which would also be called by the user
        # through the `verdi database integrity detect-duplicate-uuid` command
        deduplicate_uuids(table='db_dbnode', dry_run=False)

    def test_deduplicated_uuids(self):
        """Verify that after the migration, all expected nodes are still there with unique UUIDs."""
        # If the duplicate UUIDs were successfully fixed, the following should not raise.
        verify_uuid_uniqueness(table='db_dbnode')

        # Reload the nodes by PK and check that all UUIDs are now unique
        nodes_boolean = [self.load_node(node.pk) for node in self.nodes_boolean]
        uuids_boolean = [node.uuid for node in nodes_boolean]
        self.assertEqual(len(set(uuids_boolean)), len(nodes_boolean))

        nodes_integer = [self.load_node(node.pk) for node in self.nodes_integer]
        uuids_integer = [node.uuid for node in nodes_integer]
        self.assertEqual(len(set(uuids_integer)), len(nodes_integer))

        for node in nodes_boolean:
            self.assertEqual(utils.get_object_from_repository(node.uuid, self.file_name), self.file_content)


class TestUuidMigration(TestMigrations):

    migrate_from = '0017_drop_dbcalcstate'
    migrate_to = '0018_django_1_11'

    def setUpBeforeMigration(self):
        node = self.DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=self.default_user.id)
        node.save()

        self.node_uuid = str(node.uuid)
        self.node_id = node.id

    def test_uuid_untouched(self):
        """Verify that Node uuids remain unchanged."""
        node = self.load_node(self.node_id)
        self.assertEqual(self.node_uuid, str(node.uuid))


class TestGroupRenamingMigration(TestMigrations):

    migrate_from = '0021_dbgroup_name_to_label_type_to_type_string'
    migrate_to = '0022_dbgroup_type_string_change_content'

    def setUpBeforeMigration(self):
        # Create group
        DbGroup = self.apps.get_model('db', 'DbGroup')

        # test user group type_string: '' -> 'user'
        group_user = DbGroup(label='test_user_group', user_id=self.default_user.id, type_string='')
        group_user.save()
        self.group_user_pk = group_user.pk

        # test data.upf group type_string: 'data.upf.family' -> 'data.upf'
        group_data_upf = DbGroup(label='test_data_upf_group', user_id=self.default_user.id, type_string='data.upf.family')
        group_data_upf.save()
        self.group_data_upf_pk = group_data_upf.pk

        # test auto.import group type_string: 'aiida.import' -> 'auto.import'
        group_autoimport = DbGroup(label='test_import_group', user_id=self.default_user.id, type_string='aiida.import')
        group_autoimport.save()
        self.group_autoimport_pk = group_autoimport.pk

        # test auto.run group type_string: 'autogroup.run' -> 'auto.run'
        group_autorun = DbGroup(label='test_autorun_group', user_id=self.default_user.id, type_string='autogroup.run')
        group_autorun.save()
        self.group_autorun_pk = group_autorun.pk

    def test_group_string_update(self):
        DbGroup = self.apps.get_model('db', 'DbGroup')

        # test user group type_string: '' -> 'user'
        group_user = DbGroup.objects.get(pk=self.group_user_pk)
        self.assertEqual(group_user.type_string, 'user')

        # test data.upf group type_string: 'data.upf.family' -> 'data.upf'
        group_data_upf = DbGroup.objects.get(pk=self.group_data_upf_pk)
        self.assertEqual(group_data_upf.type_string, 'data.upf')

        # test auto.import group type_string: 'aiida.import' -> 'auto.import'
        group_autoimport = DbGroup.objects.get(pk=self.group_autoimport_pk)
        self.assertEqual(group_autoimport.type_string, 'auto.import')

        # test auto.run group type_string: 'autogroup.run' -> 'auto.run'
        group_autorun = DbGroup.objects.get(pk=self.group_autorun_pk)
        self.assertEqual(group_autorun.type_string, 'auto.run')


class TestCalcAttributeKeysMigration(TestMigrationsModelModifierV0025):

    migrate_from = '0022_dbgroup_type_string_change_content'
    migrate_to = '0023_calc_job_option_attribute_keys'

    KEY_RESOURCES_OLD = 'jobresource_params'
    KEY_RESOURCES_NEW = 'resources'
    KEY_PARSER_NAME_OLD = 'parser'
    KEY_PARSER_NAME_NEW = 'parser_name'
    KEY_PROCESS_LABEL_OLD = '_process_label'
    KEY_PROCESS_LABEL_NEW = 'process_label'
    KEY_ENVIRONMENT_VARIABLES_OLD = 'custom_environment_variables'
    KEY_ENVIRONMENT_VARIABLES_NEW = 'environment_variables'

    def setUpBeforeMigration(self):
        self.process_label = 'TestLabel'
        self.resources = {'number_machines': 1}
        self.environment_variables = {}
        self.parser_name = 'aiida.parsers:parser'

        self.node_work = self.DbNode(type='node.process.workflow.WorkflowNode.', user_id=self.default_user.id)
        self.node_work.save()
        self.set_attribute(self.node_work, self.KEY_PROCESS_LABEL_OLD, self.process_label)

        self.node_calc = self.DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=self.default_user.id)
        self.node_calc.save()
        self.set_attribute(self.node_calc, self.KEY_PROCESS_LABEL_OLD, self.process_label)
        self.set_attribute(self.node_calc, self.KEY_RESOURCES_OLD, self.resources)
        self.set_attribute(self.node_calc, self.KEY_ENVIRONMENT_VARIABLES_OLD, self.environment_variables)
        self.set_attribute(self.node_calc, self.KEY_PARSER_NAME_OLD, self.parser_name)

    def test_attribute_key_changes(self):
        """Verify that the keys are successfully changed of the affected attributes."""
        NOT_FOUND = tuple([0])

        self.assertEqual(self.get_attribute(self.node_work, self.KEY_PROCESS_LABEL_NEW), self.process_label)
        self.assertEqual(self.get_attribute(self.node_work, self.KEY_PROCESS_LABEL_OLD, default=NOT_FOUND), NOT_FOUND)

        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_PROCESS_LABEL_NEW), self.process_label)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_RESOURCES_NEW), self.resources)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_ENVIRONMENT_VARIABLES_NEW), self.environment_variables)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_PARSER_NAME_NEW), self.parser_name)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_PROCESS_LABEL_OLD, default=NOT_FOUND), NOT_FOUND)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_RESOURCES_OLD, default=NOT_FOUND), NOT_FOUND)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_ENVIRONMENT_VARIABLES_OLD, default=NOT_FOUND), NOT_FOUND)
        self.assertEqual(self.get_attribute(self.node_calc, self.KEY_PARSER_NAME_OLD, default=NOT_FOUND), NOT_FOUND)


class TestDbLogMigrationRecordCleaning(TestMigrations):

    migrate_from = '0023_calc_job_option_attribute_keys'
    migrate_to = '0024_dblog_update'

    def setUpBeforeMigration(self):
        import json
        import importlib
        from aiida.backends.sqlalchemy.utils import dumps_json

        update_024 = importlib.import_module('aiida.backends.djsite.db.migrations.0024_dblog_update')

        DbNode = self.apps.get_model('db', 'DbNode')
        DbWorkflow = self.apps.get_model('db', 'DbWorkflow')
        DbLog = self.apps.get_model('db', 'DbLog')

        # Creating the needed nodes & workflows
        calc_1 = DbNode(type="node.process.calculation.CalculationNode.", user_id=self.default_user.id)
        param = DbNode(type="data.dict.Dict.", user_id=self.default_user.id)
        leg_workf = DbWorkflow(label="Legacy WorkflowNode", user_id=self.default_user.id)
        calc_2 = DbNode(type="node.process.calculation.CalculationNode.", user_id=self.default_user.id)

        # Storing them
        calc_1.save()
        param.save()
        leg_workf.save()
        calc_2.save()

        # Creating the corresponding log records and storing them
        log_1 = DbLog(
            loggername='CalculationNode logger',
            objpk=calc_1.pk,
            objname='node.calculation.job.quantumespresso.pw.',
            message='calculation node 1',
            metadata=json.dumps({
                "msecs": 719.0849781036377,
                "objpk": calc_1.pk,
                "lineno": 350,
                "thread": 140011612940032,
                "asctime": "10/21/2018 12:39:51 PM",
                "created": 1540118391.719085,
                "levelno": 23,
                "message": "calculation node 1",
                "objname": "node.calculation.job.quantumespresso.pw.",
            }))
        log_2 = DbLog(
            loggername='something.else logger',
            objpk=param.pk,
            objname='something.else.',
            message='parameter data with log message')
        log_3 = DbLog(
            loggername='TopologicalWorkflow logger',
            objpk=leg_workf.pk,
            objname='aiida.workflows.user.topologicalworkflows.topo.TopologicalWorkflow',
            message='parameter data with log message')
        log_4 = DbLog(
            loggername='CalculationNode logger',
            objpk=calc_2.pk,
            objname='node.calculation.job.quantumespresso.pw.',
            message='calculation node 2',
            metadata=json.dumps({
                "msecs": 719.0849781036377,
                "objpk": calc_2.pk,
                "lineno": 360,
                "levelno": 23,
                "message": "calculation node 1",
                "objname": "node.calculation.job.quantumespresso.pw.",
            }))
        # Creating two more log records that don't correspond to a node
        log_5 = DbLog(
            loggername='CalculationNode logger',
            objpk=(calc_2.pk + 1000),
            objname='node.calculation.job.quantumespresso.pw.',
            message='calculation node 1000',
            metadata=json.dumps({
                "msecs": 718,
                "objpk": (calc_2.pk + 1000),
                "lineno": 361,
                "levelno": 25,
                "message": "calculation node 1000",
                "objname": "node.calculation.job.quantumespresso.pw.",
            }))
        log_6 = DbLog(
            loggername='CalculationNode logger',
            objpk=(calc_2.pk + 1001),
            objname='node.calculation.job.quantumespresso.pw.',
            message='calculation node 10001',
            metadata=json.dumps({
                "msecs": 722,
                "objpk": (calc_2.pk + 1001),
                "lineno": 362,
                "levelno": 24,
                "message": "calculation node 1001",
                "objname": "node.calculation.job.quantumespresso.pw.",
            }))

        # Storing the log records
        log_1.save()
        log_2.save()
        log_3.save()
        log_4.save()
        log_5.save()
        log_6.save()

        # Storing temporarily information needed for the check at the test
        self.to_check = dict()

        # Keeping calculation & calculation log ids
        self.to_check['CalculationNode'] = (
            calc_1.pk,
            log_1.pk,
            calc_2.pk,
            log_4.pk,
        )

        # Getting the serialized Dict logs
        param_data = DbLog.objects.filter(objpk=param.pk).filter(objname='something.else.').values(
            *update_024.values_to_export)[:1]
        serialized_param_data = dumps_json(list(param_data))
        # Getting the serialized logs for the unknown entity logs (as the export migration fuction
        # provides them) - this should coincide to the above
        serialized_unknown_exp_logs = update_024.get_serialized_unknown_entity_logs(DbLog)
        # Getting their number
        unknown_exp_logs_number = update_024.get_unknown_entity_log_number(DbLog)
        self.to_check['Dict'] = (serialized_param_data, serialized_unknown_exp_logs,
                                          unknown_exp_logs_number)

        # Getting the serialized legacy workflow logs
        leg_wf = DbLog.objects.filter(
            objpk=leg_workf.pk).filter(
            objname='aiida.workflows.user.topologicalworkflows.topo.TopologicalWorkflow').values(
            *update_024.values_to_export)[:1]
        serialized_leg_wf_logs = dumps_json(list(leg_wf))
        # Getting the serialized logs for the legacy workflow logs (as the export migration function
        # provides them) - this should coincide to the above
        serialized_leg_wf_exp_logs = update_024.get_serialized_legacy_workflow_logs(DbLog)
        eg_wf_exp_logs_number = update_024.get_legacy_workflow_log_number(DbLog)
        self.to_check['WorkflowNode'] = (serialized_leg_wf_logs, serialized_leg_wf_exp_logs, eg_wf_exp_logs_number)

        # Getting the serialized logs that don't correspond to a DbNode record
        logs_no_node = DbLog.objects.filter(
            id__in=[log_5.id, log_6.id]).values(
            *update_024.values_to_export)
        serialized_logs_no_node = dumps_json(list(logs_no_node))
        # Getting the serialized logs that don't correspond to a node (as the export migration function
        # provides them) - this should coincide to the above
        serialized_logs_exp_no_node = update_024.get_serialized_logs_with_no_nodes(DbLog, DbNode)
        logs_no_node_number = update_024.get_logs_with_no_nodes_number(DbLog, DbNode)
        self.to_check['NoNode'] = (serialized_logs_no_node, serialized_logs_exp_no_node, logs_no_node_number)

    def tearDown(self):
        """Cleaning the DbLog, DbUser, DbWorkflow and DbNode records"""
        DbUser = self.apps.get_model('db', 'DbUser')
        DbNode = self.apps.get_model('db', 'DbNode')
        DbWorkflow = self.apps.get_model('db', 'DbWorkflow')
        DbLog = self.apps.get_model('db', 'DbLog')

        DbLog.objects.all().delete()
        DbNode.objects.all().delete()  # pylint: disable=no-member
        DbWorkflow.objects.all().delete()  # pylint: disable=no-member
        DbUser.objects.all().delete()  # pylint: disable=no-member
        super(TestDbLogMigrationRecordCleaning, self).tearDown()

    def test_dblog_calculation_node(self):
        """
        Verify that after the migration there is only two log records left and verify that they corresponds to
        the CalculationNodes.
        """
        DbLog = self.apps.get_model('db', 'DbLog')

        # Check that only two log records exist
        self.assertEqual(DbLog.objects.count(), 2, "There should be two log records left")

        # Get the node id of the log record referencing the node and verify that it is the correct one
        dbnode_id_1 = DbLog.objects.filter(
            pk=self.to_check['CalculationNode'][1]).values('dbnode_id')[:1].get()['dbnode_id']
        self.assertEqual(dbnode_id_1, self.to_check['CalculationNode'][0], 'referenced node is not the expected one')
        dbnode_id_2 = DbLog.objects.filter(
            pk=self.to_check['CalculationNode'][3]).values('dbnode_id')[:1].get()['dbnode_id']
        self.assertEqual(dbnode_id_2, self.to_check['CalculationNode'][2], 'referenced node is not the expected one')

    def test_dblog_correct_export_of_logs(self):
        """
        Verify that export log methods for legacy workflows, unknown entities and log records that
        don't correspond to nodes, work as expected
        """
        import json

        self.assertEqual(self.to_check['Dict'][0], self.to_check['Dict'][1])
        self.assertEqual(self.to_check['Dict'][2], 1)

        self.assertEqual(self.to_check['WorkflowNode'][0], self.to_check['WorkflowNode'][1])
        self.assertEqual(self.to_check['WorkflowNode'][2], 1)

        self.assertEqual(sorted(list(json.loads(self.to_check['NoNode'][0])), key=lambda k: k['id']),
                         sorted(list(json.loads(self.to_check['NoNode'][1])), key=lambda k: k['id']))
        self.assertEqual(self.to_check['NoNode'][2], 2)

    def test_dblog_unique_uuids(self):
        """
        Verify that the UUIDs of the log records are unique
        """
        DbLog = self.apps.get_model('db', 'DbLog')

        l_uuids = list(_['uuid'] for _ in DbLog.objects.values('uuid'))
        s_uuids = set(l_uuids)
        self.assertEqual(len(l_uuids), len(s_uuids), "The UUIDs are not all unique.")

    def test_metadata_correctness(self):
        """
        Verify that the metadata of the remaining records don't have an objpk and objmetadata values.
        """
        import json

        DbLog = self.apps.get_model('db', 'DbLog')

        metadata = list(json.loads(_['metadata']) for _ in DbLog.objects.values('metadata'))
        # Verify that the objpk and objname are no longer part of the metadata
        for m_res in metadata:
            self.assertNotIn('objpk', m_res.keys(), "objpk should not exist any more in metadata")
            self.assertNotIn('objname', m_res.keys(), "objname should not exist any more in metadata")


class TestDbLogMigrationBackward(TestMigrations):
    """
    Check that backward migrations work also for the DbLog migration(s).
    """
    migrate_from = '0024_dblog_update'
    migrate_to = '0023_calc_job_option_attribute_keys'

    def setUpBeforeMigration(self):
        import json

        DbNode = self.apps.get_model('db', 'DbNode')
        DbLog = self.apps.get_model('db', 'DbLog')

        # Creating the needed nodes & workflows
        calc_1 = DbNode(type="node.process.calculation.CalculationNode.1", user_id=self.default_user.id)
        calc_2 = DbNode(type="node.process.calculation.CalculationNode.2", user_id=self.default_user.id)

        # Storing them
        calc_1.save()
        calc_2.save()

        # Creating the corresponding log records and storing them
        log_1 = DbLog(
            loggername='CalculationNode logger',
            dbnode_id=calc_1.pk,
            message='calculation node 1',
            metadata=json.dumps({
                "msecs": 719.0849781036377,
                "lineno": 350,
                "thread": 140011612940032,
                "asctime": "10/21/2018 12:39:51 PM",
                "created": 1540118391.719085,
                "levelno": 23,
                "message": "calculation node 1",
            }))
        log_2 = DbLog(
            loggername='CalculationNode logger',
            dbnode_id=calc_2.pk,
            message='calculation node 2',
            metadata=json.dumps({
                "msecs": 719.0849781036377,
                "lineno": 360,
                "levelno": 23,
                "message": "calculation node 1",
            }))

        # Storing the log records
        log_1.save()
        log_2.save()

        # Keeping what is needed to be verified at the test
        self.to_check = dict()
        self.to_check[log_1.pk] = (log_1.dbnode_id, calc_1.type)
        self.to_check[log_2.pk] = (log_2.dbnode_id, calc_2.type)

    def test_objpk_objname(self):
        """
        This test verifies that the objpk and objname have the right values
        after a forward and a backward migration.
        """
        import json
        DbLog = self.apps.get_model('db', 'DbLog')

        # Check that only two log records exist with the correct objpk objname
        for log_pk in self.to_check.keys():
            log_entry = DbLog.objects.filter(pk=log_pk)[:1].get()
            log_dbnode_id, type = self.to_check[log_pk]
            self.assertEqual(log_dbnode_id, log_entry.objpk,
                             "The dbnode_id ({}) of the 0024 schema version should be identical to the objpk ({}) of "
                             "the 0023 schema version.".format(log_dbnode_id, log_entry.objpk))
            self.assertEqual(type, log_entry.objname,
                             "The type ({}) of the linked node of the 0024 schema version should be identical to the "
                             "objname ({}) of the 0023 schema version.".format(type, log_entry.objname))
            self.assertEqual(log_dbnode_id, json.loads(log_entry.metadata)['objpk'],
                             "The dbnode_id ({}) of the 0024 schema version should be identical to the objpk ({}) of "
                             "the 0023 schema version stored in the metadata.".format(
                                 log_dbnode_id, json.loads(log_entry.metadata)['objpk']))
            self.assertEqual(type, json.loads(log_entry.metadata)['objname'],
                             "The type ({}) of the linked node of the 0024 schema version should be identical to the "
                             "objname ({}) of the 0023 schema version stored in the metadata.".format(
                                 type, json.loads(log_entry.metadata)['objname']))


class TestDataMoveWithinNodeMigration(TestMigrations):

    migrate_from = '0024_dblog_update'
    migrate_to = '0025_move_data_within_node_module'

    def setUpBeforeMigration(self):
        self.node_calc = self.DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=self.default_user.id)
        self.node_data = self.DbNode(type='data.int.Int.', user_id=self.default_user.id)
        self.node_calc.save()
        self.node_data.save()

    def test_data_type_string(self):
        """Verify that type string of the Data node was successfully adapted."""
        node_calc = self.load_node(self.node_calc.id)
        node_data = self.load_node(self.node_data.id)
        self.assertEqual(node_data.type, 'node.data.int.Int.')
        self.assertEqual(node_calc.type, 'node.process.calculation.calcjob.CalcJobNode.')


class TestTrajectoryDataMigration(TestMigrationsModelModifierV0025):

    migrate_from = '0025_move_data_within_node_module'
    migrate_to = '0027_delete_trajectory_symbols_array'

    stepids = numpy.array([60, 70])
    times = stepids * 0.01
    positions = numpy.array(
        [[[0., 0., 0.], [0.5, 0.5, 0.5], [1.5, 1.5, 1.5]], [[0., 0., 0.], [0.5, 0.5, 0.5], [1.5, 1.5, 1.5]]])
    velocities = numpy.array(
        [[[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]], [[0.5, 0.5, 0.5], [0.5, 0.5, 0.5], [-0.5, -0.5, -0.5]]])
    cells = numpy.array([[[2., 0., 0.], [0., 2., 0.], [0., 0., 2.]], [[3., 0., 0.], [0., 3., 0.], [0., 0., 3.]]])

    def setUpBeforeMigration(self):
        symbols = numpy.array(['H', 'O', 'C'])
        self.node = self.DbNode(type='node.data.array.trajectory.TrajectoryData.', user_id=self.default_user.id)
        self.node.save()

        self.set_node_array(self.node, 'steps', self.stepids)
        self.set_node_array(self.node, 'cells', self.cells)
        self.set_node_array(self.node, 'symbols', symbols)
        self.set_node_array(self.node, 'positions', self.positions)
        self.set_node_array(self.node, 'times', self.times)
        self.set_node_array(self.node, 'velocities', self.velocities)

    def test_trajectory_symbols(self):
        node = self.load_node(self.node.id)
        self.assertSequenceEqual(self.get_attribute(node, 'symbols'), ['H', 'O', 'C'])
        self.assertSequenceEqual(self.get_node_array(node, 'velocities').tolist(), self.velocities.tolist())
        self.assertSequenceEqual(self.get_node_array(node, 'positions').tolist(), self.positions.tolist())
        with self.assertRaises(IOError):
            self.get_node_array(node, 'symbols')


class TestNodePrefixRemovalMigration(TestMigrations):

    migrate_from = '0027_delete_trajectory_symbols_array'
    migrate_to = '0028_remove_node_prefix'

    def setUpBeforeMigration(self):
        self.node_calc = self.DbNode(type='node.process.calculation.calcjob.CalcJobNode.', user_id=self.default_user.id)
        self.node_data = self.DbNode(type='node.data.int.Int.', user_id=self.default_user.id)
        self.node_calc.save()
        self.node_data.save()

    def test_data_node_type_string(self):
        """Verify that type string of the nodes was successfully adapted."""
        node_calc = self.load_node(self.node_calc.id)
        node_data = self.load_node(self.node_data.id)

        self.assertEqual(node_data.type, 'data.int.Int.')
        self.assertEqual(node_calc.type, 'process.calculation.calcjob.CalcJobNode.')


class TestParameterDataToDictMigration(TestMigrations):

    migrate_from = '0028_remove_node_prefix'
    migrate_to = '0029_rename_parameter_data_to_dict'

    def setUpBeforeMigration(self):
        self.node = self.DbNode(type='data.parameter.ParameterData.', user_id=self.default_user.id)
        self.node.save()

    def test_data_node_type_string(self):
        """Verify that type string of the nodes was successfully adapted."""
        node = self.load_node(self.node.id)
        self.assertEqual(node.type, 'data.dict.Dict.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Tests for nodes, attributes and links
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from aiida.backends.testbase import AiidaTestCase
from aiida.common.links import LinkType
from aiida.orm import Data, Node
from aiida.orm import CalculationNode


class TestNodeBasicDjango(AiidaTestCase):
    def test_replace_extras_2(self):
        """
        This is a Django specific test which checks (manually) that,
        when replacing list and dict with objects that have no deepness,
        no junk is left in the DB (i.e., no 'dict.a', 'list.3.h', ...
        """
        from aiida.backends.djsite.db.models import DbExtra

        a = Data().store()
        extras_to_set = {
            'bool': True,
            'integer': 12,
            'float': 26.2,
            'string': "a string",
            'dict': {"a": "b",
                     "sublist": [1, 2, 3],
                     "subdict": {
                         "c": "d"}},
            'list': [1, True, "ggg", {'h': 'j'}, [9, 8, 7]],
        }

        # I redefine the keys with more complicated data, and
        # changing the data type too
        new_extras = {
            'bool': 12,
            'integer': [2, [3], 'a'],
            'float': {'n': 'm', 'x': [1, 'r', {}]},
            'string': True,
            'dict': 'text',
            'list': 66.3,
        }

        for k, v in extras_to_set.items():
            a.set_extra(k, v)

        for k, v in new_extras.items():
            # I delete one by one the keys and check if the operation is
            # performed correctly
            a.set_extra(k, v)

        # I update extras_to_set with the new entries, and do the comparison
        # again
        extras_to_set.update(new_extras)

        # Check (manually) that, when replacing list and dict with objects
        # that have no deepness, no junk is left in the DB (i.e., no
        # 'dict.a', 'list.3.h', ...
        self.assertEquals(len(DbExtra.objects.filter(
            dbnode=a.backend_entity.dbmodel, key__startswith=('list' + DbExtra._sep))), 0)
        self.assertEquals(len(DbExtra.objects.filter(
            dbnode=a.backend_entity.dbmodel, key__startswith=('dict' + DbExtra._sep))), 0)

    def test_attrs_and_extras_wrong_keyname(self):
        """
        Attribute keys cannot include the separator symbol in the key
        """
        from aiida.backends.djsite.db.models import DbAttributeBaseClass
        from aiida.common.exceptions import ModificationNotAllowed, ValidationError

        separator = DbAttributeBaseClass._sep

        a = Data().store()

        with self.assertRaises(ModificationNotAllowed):
            # Cannot change an attribute on a stored node
            a.set_attribute('name' + separator, 'blablabla')

        with self.assertRaises(ValidationError):
            # Cannot change an attribute on a stored node
            a.set_attribute('name' + separator, 'blablabla', stored_check=False)

        with self.assertRaises(ValidationError):
            # Passing an attribute key separator directly in the key is not allowed
            a.set_extra('bool' + separator, 'blablabla')

    def test_settings(self):
        """
        Test the settings table (similar to Attributes, but without the key.
        """
        from aiida.backends.djsite.db import models
        from django.db import IntegrityError, transaction

        models.DbSetting.set_value(key='pippo', value=[1, 2, 3])

        s1 = models.DbSetting.objects.get(key='pippo')

        self.assertEqual(s1.getvalue(), [1, 2, 3])

        s2 = models.DbSetting(key='pippo')

        sid = transaction.savepoint()
        with self.assertRaises(IntegrityError):
            # same name...
            s2.save()
        transaction.savepoint_rollback(sid)

        # Should replace pippo
        models.DbSetting.set_value(key='pippo', value="a")
        s1 = models.DbSetting.objects.get(key='pippo')

        self.assertEqual(s1.getvalue(), "a")

    def test_load_nodes(self):
        """
        """
        from aiida.orm import load_node

        a = Data()
        a.store()
        self.assertEquals(a.pk, load_node(pk=a.pk).pk)
        self.assertEquals(a.pk, load_node(uuid=a.uuid).pk)

        with self.assertRaises(ValueError):
            load_node(identifier=a.pk, pk=a.pk)
        with self.assertRaises(ValueError):
            load_node(pk=a.pk, uuid=a.uuid)
        with self.assertRaises(TypeError):
            load_node(pk=a.uuid)
        with self.assertRaises(TypeError):
            load_node(uuid=a.pk)
        with self.assertRaises(ValueError):
            load_node()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for `verdi setup`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from click.testing import CliRunner

from aiida.backends import settings
from aiida.backends.testbase import AiidaTestCase
from aiida.backends.tests.utils.configuration import with_temporary_config_instance
from aiida.cmdline.commands import cmd_setup
from aiida.manage.configuration import get_config


class TestVerdiSetup(AiidaTestCase):
    """Tests for `verdi setup`."""

    def setUp(self):
        """Create runner object to run tests."""
        self.cli_runner = CliRunner()

    @with_temporary_config_instance
    def test_non_interactive_override(self):
        """Tests that existing profile can be overridden in non-interactive mode with the `--force` option."""
        config = get_config()
        profile = config.current_profile

        option_string = '--backend={backend} --email={email} --repository={repo} --db-host={db_host} ' \
                        '--db-port={db_port} --db-name={db_name} --db-username={db_user} --db-password={db_pass} ' \
                        '--first-name={first} --last-name={last} --institution={institution}'

        profile_settings = {
            'email': profile.dictionary['default_user_email'],
            'backend': profile.dictionary['AIIDADB_BACKEND'],
            'db_host': profile.dictionary['AIIDADB_HOST'],
            'db_port': profile.dictionary['AIIDADB_PORT'],
            'db_name': profile.dictionary['AIIDADB_NAME'],
            'db_user': profile.dictionary['AIIDADB_USER'],
            'db_pass': profile.dictionary['AIIDADB_PASS'],
            'repo': '/tmp',
            'first': 'nim',
            'last': 'porte',
            'institution': 'quoi',
        }

        options = ['--non-interactive'] + option_string.format(**profile_settings).split()

        # Without profile name should raise exception
        result = self.cli_runner.invoke(cmd_setup.setup, options)
        self.assertIsNotNone(result.exception)
        self.assertIn('Missing argument', result.output)
        self.assertIn('PROFILE_NAME', result.output)

        # With existing profile name should raise exception
        options.append(profile.name)
        result = self.cli_runner.invoke(cmd_setup.setup, options)
        self.assertIsNotNone(result.exception, result.output)

        # Need to set the current profile to None again, otherwise it will complain that `verdi` is using `-p` flag
        settings.AIIDADB_PROFILE = None

        # Adding the `--force` option should allow it
        options.append('--force')
        result = self.cli_runner.invoke(cmd_setup.setup, options)
        self.assertIsNone(result.exception, result.output)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for `verdi`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from click.testing import CliRunner

from aiida import get_version
from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands import cmd_verdi


class TestVerdi(AiidaTestCase):
    """Tests for `verdi run`."""

    def setUp(self):
        super(TestVerdi, self).setUp()
        self.cli_runner = CliRunner()

    def test_verdi_version(self):
        """Regression test for #2238: verify that `verdi --version` prints the current version"""
        result = self.cli_runner.invoke(cmd_verdi.verdi, ['--version'])
        self.assertIsNone(result.exception, result.output)
        self.assertIn(get_version(), result.output)

    def test_verdi_with_empty_profile_list(self):
        # pylint: disable=protected-access
        """Test for #2424: verify that verdi remains operable even if profile list is empty"""
        from aiida.manage.configuration import CONFIG

        # Run verdi command with updated CONFIG featuring an empty profile list
        config_dict = dict(CONFIG._dictionary)
        CONFIG._dictionary[CONFIG.KEY_PROFILES] = {}
        result = self.cli_runner.invoke(cmd_verdi.verdi, [])
        self.assertIsNone(result.exception, result.output)

        # Restore original CONFIG contents
        CONFIG._dictionary = config_dict



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module that defines the configuration file of an AiiDA instance and functions to create and load it."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import io
import os
import shutil

from aiida.common import json

from .migrations import CURRENT_CONFIG_VERSION, OLDEST_COMPATIBLE_CONFIG_VERSION
from .options import get_option, parse_option, NO_DEFAULT
from .profile import Profile
from .settings import DEFAULT_UMASK, DEFAULT_CONFIG_INDENT_SIZE

__all__ = ('Config',)


class Config(object):  # pylint: disable=useless-object-inheritance
    """Object that represents the configuration file of an AiiDA instance."""

    KEY_VERSION = 'CONFIG_VERSION'
    KEY_VERSION_CURRENT = 'CURRENT'
    KEY_VERSION_OLDEST_COMPATIBLE = 'OLDEST_COMPATIBLE'
    KEY_DEFAULT_PROFILE = 'default_profile'
    KEY_PROFILES = 'profiles'

    def __init__(self, filepath, dictionary):
        """Instantiate a configuration object from a configuration dictionary and its filepath.

        If an empty dictionary is passed, the constructor will create the skeleton configuration dictionary.

        :param filepath: the absolute filepath of the configuration file
        :param dictionary: the content of the configuration file in dictionary form
        """
        if not dictionary:
            # Construct the default configuration template
            dictionary = {
                self.KEY_VERSION: {
                    self.KEY_VERSION_CURRENT: CURRENT_CONFIG_VERSION,
                    self.KEY_VERSION_OLDEST_COMPATIBLE: OLDEST_COMPATIBLE_CONFIG_VERSION,
                },
                self.KEY_PROFILES: {},
            }

        self._filepath = filepath
        self._dictionary = dictionary

    def __eq__(self, other):
        """Two configurations are considered equal, when their dictionaries are equal."""
        return self.dictionary == other.dictionary

    def __ne__(self, other):
        """Two configurations are considered unequal, when their dictionaries are unequal."""
        return self.dictionary != other.dictionary

    @property
    def dictionary(self):
        return self._dictionary

    @property
    def version(self):
        return self._version_settings.get(self.KEY_VERSION_CURRENT, 0)

    @version.setter
    def version(self, version):
        self._version_settings[self.KEY_VERSION_CURRENT] = version

    @property
    def version_oldest_compatible(self):
        return self._version_settings.get(self.KEY_VERSION_OLDEST_COMPATIBLE, 0)

    @version_oldest_compatible.setter
    def version_oldest_compatible(self, version_oldest_compatible):
        self._version_settings[self.KEY_VERSION_OLDEST_COMPATIBLE] = version_oldest_compatible

    @property
    def _version_settings(self):
        return self.dictionary.setdefault(self.KEY_VERSION, {})

    @property
    def filepath(self):
        return self._filepath

    @property
    def dirpath(self):
        return os.path.dirname(self.filepath)

    @property
    def default_profile_name(self):
        """Return the default profile name.

        :return: the default profile name or None if not defined
        """
        try:
            return self.dictionary[self.KEY_DEFAULT_PROFILE]
        except KeyError:
            return None

    @property
    def current_profile(self):
        """Return the currently loaded profile.

        :return: the current profile or None if not defined
        """
        from aiida.backends import settings

        current_profile_name = settings.AIIDADB_PROFILE

        if current_profile_name:
            return self.get_profile(current_profile_name)

        return None

    @property
    def profile_names(self):
        """Return the list of profile names.

        :return: list of profile names
        """
        try:
            profiles = self.dictionary[self.KEY_PROFILES]
        except KeyError:
            return []
        else:
            return list(profiles)

    @property
    def profiles(self):
        """Return the list of profiles.

        :return: the profiles
        :rtype: list of `Profile` instances
        """
        try:
            profiles = self.dictionary[self.KEY_PROFILES]
        except KeyError:
            return []
        else:
            return [Profile(name, profile) for name, profile in profiles.items()]

    def validate_profile(self, name):
        """Validate that a profile exists.

        :param name: name of the profile:
        :raises aiida.common.ProfileConfigurationError: if the name is not found in the configuration file
        """
        from aiida.common import exceptions

        if name not in self.profile_names:
            raise exceptions.ProfileConfigurationError('profile `{}` does not exist'.format(name))

    def get_profile(self, name=None):
        """Return the profile for the given name or the default one if not specified.

        :return: the profile instance or None if it does not exist
        :raises aiida.common.ProfileConfigurationError: if the name is not found in the configuration file
        """
        from aiida.common import exceptions

        if not name and not self.default_profile_name:
            raise exceptions.ProfileConfigurationError('no default profile defined')

        if not name:
            name = self.default_profile_name

        self.validate_profile(name)

        return Profile(name, self.dictionary[self.KEY_PROFILES][name])

    def _get_profile_dictionary(self, name):
        """Return the internal profile dictionary for the given name or the default one if not specified.

        :return: the profile instance or None if it does not exist
        :raises aiida.common.ProfileConfigurationError: if the name is not found in the configuration file
        """
        self.validate_profile(name)
        return self.dictionary[self.KEY_PROFILES][name]

    def add_profile(self, name, profile):
        """Add a profile to the configuration.

        :param name: the name of the profile to remove
        :param profile: the profile configuration dictionary
        :return: self
        """
        self.dictionary[self.KEY_PROFILES][name] = profile
        return self

    def update_profile(self, profile):
        """Update a profile in the configuration.

        :param profile: the profile instance to update
        :return: self
        """
        self.dictionary[self.KEY_PROFILES][profile.name] = profile.dictionary
        return self

    def remove_profile(self, name):
        """Remove a profile from the configuration.

        :param name: the name of the profile to remove
        :raises aiida.common.ProfileConfigurationError: if the given profile does not exist
        :return: self
        """
        self.validate_profile(name)
        self.dictionary[self.KEY_PROFILES].pop(name)
        return self

    def set_default_profile(self, name, overwrite=False):
        """Set the given profile as the new default.

        :param name: name of the profile to set as new default
        :param overwrite: when True, set the profile as the new default even if a default profile is already defined
        :raises aiida.common.ProfileConfigurationError: if the given profile does not exist
        :return: self
        """
        if self.default_profile_name and not overwrite:
            return self

        self.validate_profile(name)
        self.dictionary[self.KEY_DEFAULT_PROFILE] = name
        return self

    def option_set(self, option_name, option_value, scope=None):
        """Set a configuration option.

        :param option_name: the name of the configuration option
        :param option_value: the option value
        :param scope: set the option for this profile or globally if not specified
        """
        option, parsed_value = parse_option(option_name, option_value)

        if scope is not None:
            dictionary = self._get_profile_dictionary(scope)
        else:
            dictionary = self.dictionary

        if parsed_value:
            dictionary[option.key] = parsed_value
        elif option.default is not NO_DEFAULT:
            dictionary[option.key] = option.default
        else:
            pass

    def option_unset(self, option_name, scope=None):
        """Unset a configuration option.

        :param option_name: the name of the configuration option
        :param scope: unset the option for this profile or globally if not specified
        """
        option = get_option(option_name)

        if scope is not None:
            dictionary = self._get_profile_dictionary(scope)
        else:
            dictionary = self.dictionary

        dictionary.pop(option.key, None)

    def option_get(self, option_name, scope=None, default=True):
        """Get a configuration option.

        :param option_name: the name of the configuration option
        :param scope: get the option for this profile or globally if not specified
        :param default: boolean, If True will return the option default, even if not defined within the given scope
        :return: the option value or None if not set for the given scope
        """
        option = get_option(option_name)

        if scope is not None:
            dictionary = self.get_profile(scope).dictionary
        else:
            dictionary = self.dictionary

        return dictionary.get(option.key, option.default if default else None)

    def store(self):
        """Write the current config to file."""
        self._backup()

        umask = os.umask(DEFAULT_UMASK)

        try:
            with io.open(self.filepath, 'wb') as handle:
                json.dump(self.dictionary, handle, indent=DEFAULT_CONFIG_INDENT_SIZE)
        finally:
            os.umask(umask)

        return self

    def _backup(self):
        """Create a backup of the current config as it exists on disk."""
        if not os.path.isfile(self.filepath):
            return

        umask = os.umask(DEFAULT_UMASK)

        try:
            shutil.copy(self.filepath, self.filepath + '~')
        finally:
            os.umask(umask)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import six

from sqlalchemy.schema import Column
from sqlalchemy.types import Integer, String, Boolean, Text
from sqlalchemy.orm import relationship, backref
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

from sqlalchemy.dialects.postgresql import UUID, JSONB

from aiida.backends.sqlalchemy.models.base import Base
from aiida.common.utils import get_new_uuid

from aiida.common.exceptions import NotExistent, DbContentError, ConfigurationError


class DbComputer(Base):
    __tablename__ = "db_dbcomputer"

    id = Column(Integer, primary_key=True)

    uuid = Column(UUID(as_uuid=True), default=get_new_uuid, unique=True)
    name = Column(String(255), unique=True, nullable=False)
    hostname = Column(String(255))

    description = Column(Text, nullable=True)
    enabled = Column(Boolean)

    transport_type = Column(String(255))
    scheduler_type = Column(String(255))

    transport_params = Column(JSONB)
    _metadata = Column('metadata', JSONB)

    def __init__(self, *args, **kwargs):
        self.enabled = True
        self._metadata = {}
        self.transport_params = {}
        # TODO SP: it's supposed to be nullable, but there is a NOT NULL
        # constraint inside the DB.
        self.description = ""

        super(DbComputer, self).__init__(*args, **kwargs)

    @classmethod
    def get_dbcomputer(cls, computer):
        """
        Return a DbComputer from its name (or from another Computer or DbComputer instance)
        """

        from aiida.orm.computers import Computer
        if isinstance(computer, six.string_types):
            try:
                dbcomputer = cls.session.query(cls).filter(cls.name == computer).one()
            except NoResultFound:
                raise NotExistent("No computer found in the table of computers with "
                                  "the given name '{}'".format(computer))
            except MultipleResultsFound:
                raise DbContentError("There is more than one computer with name '{}', "
                                     "pass a Computer instance".format(computer))
        elif isinstance(computer, int):
            try:
                dbcomputer = cls.session.query(cls).filter(cls.id == computer).one()
            except NoResultFound:
                raise NotExistent("No computer found in the table of computers with "
                                  "the given id '{}'".format(computer))
        elif isinstance(computer, DbComputer):
            if computer.id is None:
                raise ValueError("The computer instance you are passing has not been stored yet")
            dbcomputer = computer
        elif isinstance(computer, Computer):
            if computer.dbcomputer.id is None:
                raise ValueError("The computer instance you are passing has not been stored yet")
            dbcomputer = computer.dbcomputer
        else:
            raise TypeError(
                "Pass either a computer name, a DbComputer SQLAlchemy instance, a Computer id or a Computer object")
        return dbcomputer

    @property
    def pk(self):
        return self.id

    def __str__(self):
        if self.enabled:
            return "{} ({})".format(self.name, self.hostname)
        else:
            return "{} ({}) [DISABLED]".format(self.name, self.hostname)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
utilities for:

* managing the registry cache folder
* downloading json files
* pickling to the registry cache folder
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import io


def registry_cache_folder_name():
    """
    return the name of the subfolder of aiida_dir where registry caches are stored.
    """
    return 'plugin_registry_cache'


def registry_cache_folder_path():
    """
    return the fully resolved path to the cache folder
    """
    from os import path as osp
    from aiida.manage.configuration import get_config
    config = get_config()
    cache_dir = registry_cache_folder_name()
    return osp.join(config.dirpath, cache_dir)


def registry_cache_folder_exists():
    """
    return True if the cache folder exists, False if not
    """
    from os import path as osp
    cache_dir = registry_cache_folder_path()
    return osp.isdir(cache_dir)


def safe_create_registry_cache_folder():
    """
    creates the registry cache folder if it does not exist
    """
    from os import mkdir
    if not registry_cache_folder_exists():
        cache_dir = registry_cache_folder_path()
        mkdir(cache_dir)


def pickle_to_registry_cache_folder(obj, fname):
    """
    pickles a python object to the registry cache folder
    """
    from six.moves.cPickle import dump as pdump
    from os import path as osp
    safe_create_registry_cache_folder()
    cache_dir = registry_cache_folder_path()
    fpath = osp.join(cache_dir, fname)
    with io.open(fpath, 'w', encoding='utf8') as cache_file:
        pdump(obj, cache_file)


def unpickle_from_registry_cache_folder(fname):
    """
    looks for fname in the registry cache folder and tries to unpickle from it
    """
    from six.moves.cPickle import load as pload
    from os import path as osp
    cache_dir = registry_cache_folder_path()
    fpath = osp.join(cache_dir, fname)
    with io.open(fpath, 'r', encoding='utf8') as cache:
        return pload(cache)


def load_json_from_url(url, errorhandler=None):
    """
    downloads a json file and returns the corresponding python dict
    """
    import requests
    reply = requests.get(url)
    res = None
    try:
        res = reply.json()
    except Exception as e:
        if errorhandler:
            res = errorhandler(e)
        else:
            raise e
    return res


def value_error_msg(e):
    msg = 'The AiiDA plugin registry seems to be temporarily unavailable.'
    msg += ' Please try again later. If the problem persists,'
    msg += ' look at github.com/aiidateam/aiida-registry and file an issue'
    msg += ' if there is none yet.'
    return msg



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Module containing utilities and classes relating to job calculations running
on systems that require transport.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import contextlib
from functools import partial
import time

from six import iteritems, itervalues
from tornado import concurrent, gen

from aiida import schedulers
from aiida.common import exceptions
from ...utils import RefObjectStore

__all__ = ('JobsList', 'JobManager')


class JobsList(object):  # pylint: disable=useless-object-inheritance
    """
    A list of submitted jobs on a machine connected to by transport based on the
    authorisation information.
    """

    def __init__(self, authinfo, transport_queue):
        """
        :param authinfo: The authinfo used to check the jobs list
        :type authinfo: :class:`aiida.orm.AuthInfo`
        :param transport_queue: A transport queue
        :type: :class:`aiida.engine.transports.TransportQueue`
        """
        self._authinfo = authinfo
        self._transport_queue = transport_queue
        self._loop = transport_queue.loop()

        self._jobs_cache = {}
        self._last_updated = None  # type: float
        self._job_update_requests = {}  # Mapping: {job_id: Future}
        self._update_handle = None

    def get_minimum_update_interval(self):
        """
        Get the minimum interval that can be expected between updates of the list
        :return: The minimum interval
        :rtype: float
        """
        return self._authinfo.computer.get_minimum_job_poll_interval()

    def get_last_updated(self):
        """
        Get the timestamp of when the list was last updated as produced by `time.time()`

        :return: The last update point
        :rtype: float
        """
        return self._last_updated

    @gen.coroutine
    def _get_jobs_from_scheduler(self):
        """
        Get the current jobs list from the scheduler

        :return: A dictionary of {job_id: job info}
        :rtype: dict
        """
        with self._transport_queue.request_transport(self._authinfo) as request:
            transport = yield request

            scheduler = self._authinfo.computer.get_scheduler()
            scheduler.set_transport(transport)

            kwargs = {'as_dict': True}
            if scheduler.get_feature('can_query_by_user'):
                kwargs['user'] = "$USER"
            else:
                kwargs['jobs'] = self._get_jobs_with_scheduler()

            scheduler_response = scheduler.get_jobs(**kwargs)
            jobs_cache = {}

            for job_id, job_info in iteritems(scheduler_response):
                # If the job is done then get detailed job information
                detailed_job_info = None
                if job_info.job_state == schedulers.JobState.DONE:
                    try:
                        detailed_job_info = scheduler.get_detailed_jobinfo(job_id)
                    except exceptions.FeatureNotAvailable:
                        detailed_job_info = 'This scheduler does not implement get_detailed_jobinfo'

                job_info.detailedJobinfo = detailed_job_info
                jobs_cache[job_id] = job_info

            raise gen.Return(jobs_cache)

    @gen.coroutine
    def _update_job_info(self):
        """
        Update all of the job information objects for a given authinfo, that is to say for
        all the jobs on a particular machine for a particular user.

        This will set the futures for all pending update requests where the corresponding job
        has a new status compared to the last update.
        """
        try:
            if not self._update_requests_outstanding():
                return

            # Update our cache of the job states
            self._jobs_cache = yield self._get_jobs_from_scheduler()
        except Exception as exception:
            # Set the exception on all the update futures
            for future in itervalues(self._job_update_requests):
                if not future.done():
                    future.set_exception(exception)
            raise
        else:
            for job_id, future in iteritems(self._job_update_requests):
                if not future.done():
                    future.set_result(self._jobs_cache.get(job_id, None))
        finally:
            self._job_update_requests = {}

    @contextlib.contextmanager
    def request_job_info_update(self, job_id):
        """
        Request job info about a job when it next changes it's job state.  If the job is not
        found in the jobs list at the update the future will resolve to None.

        :param job_id: The job identifier
        :return: A future that will resolve to a JobInfo object when the job changes state
        """
        # Get or create the future
        request = self._job_update_requests.setdefault(job_id, concurrent.Future())
        assert not request.done(), "The future should be no be in the done state"

        try:
            self._ensure_updating()
            yield request
        finally:
            pass

    def _ensure_updating(self):
        """
        Ensure that we are updating the job list from the remote resource.
        This will automatically stop if there are no outstanding requests.
        """

        @gen.coroutine
        def updating():
            """ Do the actual update, stop if not requests left """
            yield self._update_job_info()
            # Any outstanding requests?
            if self._update_requests_outstanding():
                self._update_handle = self._loop.call_later(self._get_next_update_delay(), updating)
            else:
                self._update_handle = None

        # Check if we're already updating
        if self._update_handle is None:
            self._update_handle = self._loop.call_later(self._get_next_update_delay(), updating)

    @staticmethod
    def _has_job_state_changed(old, new):
        """
        :type old: :class:`aiida.schedulers.JobInfo`
        :type new: :class:`aiida.schedulers.JobInfo`
        :rtype: bool
        """
        if old is None and new is None:
            return False

        if old is None or new is None:
            # One is None and the other isn't
            return True

        return old.job_state != new.job_state or old.job_substate != new.job_substate

    def _get_next_update_delay(self):
        """
        Calculate when we are next allowed to call the scheduler get jobs command
        based on when we last called it, how long has elapsed and the minimum given
        update interval.

        :return: The delay (in seconds) for when it's safe to call the get jobs command
        :rtype: float
        """
        if self._last_updated is None:
            # Never updated, so do it straight away
            return 0.

        # Make sure to actually 'get' it here, so that if the user changed it
        # between times we use the current value
        minimum_interval = self._authinfo.computer.get_minimum_job_poll_interval()
        elapsed = time.time() - self._last_updated

        return max(minimum_interval - elapsed, 0.)

    def _update_requests_outstanding(self):
        return any(not request.done() for request in itervalues(self._job_update_requests))

    def _get_jobs_with_scheduler(self):
        """
        Get all the jobs that are currently with scheduler for this authinfo

        :return: the list of jobs with the scheduler
        :rtype: list
        """
        return [str(job_id) for job_id, _ in self._job_update_requests.items()]


class JobManager(object):  # pylint: disable=useless-object-inheritance
    """
    A manager for jobs on a (usually) remote resource such as a supercomputer
    """

    def __init__(self, transport_queue):
        self._transport_queue = transport_queue
        self._job_lists = RefObjectStore()

    @contextlib.contextmanager
    def request_job_info_update(self, authinfo, job_id):
        """
        Get a future that will resolve to information about a given job.  This is a context
        manager so that if the user leaves the context the request is automatically cancelled.

        :return: A tuple containing the JobInfo object and detailed job info.  Both can be None.
        :rtype: :class:`tornado.concurrent.Future`
        """
        # Define a way to create a JobsList if needed
        create = partial(JobsList, authinfo, self._transport_queue)

        with self._job_lists.get(authinfo.id, create) as job_list:
            with job_list.request_job_info_update(job_id) as request:
                try:
                    yield request
                finally:
                    if not request.done():
                        request.cancel()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""SqlAlchemy implementations for the AuthInfo entity and collection."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# pylint: disable=import-error,no-name-in-module
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound

from aiida.backends.sqlalchemy import get_scoped_session
from aiida.backends.sqlalchemy.models.authinfo import DbAuthInfo
from aiida.common import exceptions
from aiida.common.lang import type_check
from aiida.orm.implementation.authinfos import BackendAuthInfo, BackendAuthInfoCollection

from . import entities
from . import utils


class SqlaAuthInfo(entities.SqlaModelEntity[DbAuthInfo], BackendAuthInfo):
    """AuthInfo implementation for SQLAlchemy."""

    MODEL_CLASS = DbAuthInfo

    def __init__(self, backend, computer, user):
        """
        Construct an SqlaAuthInfo

        :param computer: a Computer instance
        :param user: a User instance
        :return: an AuthInfo object associated with the given computer and user
        """
        from . import computers
        from . import users
        super(SqlaAuthInfo, self).__init__(backend)
        type_check(user, users.SqlaUser)
        type_check(computer, computers.SqlaComputer)
        self._dbmodel = utils.ModelWrapper(DbAuthInfo(dbcomputer=computer.dbmodel, aiidauser=user.dbmodel))

    @property
    def dbauthinfo(self):
        return self._dbmodel._model  # pylint: disable=protected-access

    @property
    def is_stored(self):
        """
        Return whether the AuthInfo is stored

        :return: True if stored, False otherwise
        """
        return self._dbmodel.is_saved()

    @property
    def id(self):
        return self._dbmodel.id

    @property
    def enabled(self):
        return self._dbmodel.enabled

    @enabled.setter
    def enabled(self, enabled):
        self._dbmodel.enabled = enabled

    @property
    def computer(self):
        return self._backend.computers.from_dbmodel(self._dbmodel.dbcomputer)

    @property
    def user(self):
        return self._backend.users.from_dbmodel(self._dbmodel.aiidauser)

    def get_auth_params(self):
        """
        Get the auth_params dictionary from the DB

        :return: a dictionary
        """
        return self._dbmodel.auth_params

    def set_auth_params(self, auth_params):
        """
        Replace the auth_params dictionary in the DB with the provided dictionary
        """
        # Raises ValueError if data is not JSON-serializable
        self._dbmodel.auth_params = auth_params

    def get_metadata(self):
        """
        Get the metadata dictionary from the DB

        :return: a dictionary
        """
        return self._dbmodel._metadata  # pylint: disable=protected-access

    def set_metadata(self, metadata):
        """
        Replace the metadata dictionary in the DB with the provided dictionary
        """
        # Raises ValueError if data is not JSON-serializable
        self._dbmodel._metadata = metadata  # pylint: disable=protected-access


class SqlaAuthInfoCollection(BackendAuthInfoCollection):
    """Collection of AuthInfo instances."""

    ENTITY_CLASS = SqlaAuthInfo

    def get(self, computer, user):
        """
        Return a SqlaAuthInfo given a computer and a user

        :param computer: a Computer instance
        :param user: a User instance
        :return: an AuthInfo object associated with the given computer and user
        :raise aiida.common.NotExistent: if the user is not configured to use computer
        :raise sqlalchemy.orm.exc.MultipleResultsFound: if the user is configured
             more than once to use the computer! Should never happen
        """
        session = get_scoped_session()

        try:
            authinfo = session.query(DbAuthInfo).filter_by(
                dbcomputer_id=computer.id,
                aiidauser_id=user.id,
            ).one()

            return self.from_dbmodel(authinfo)
        except NoResultFound:
            raise exceptions.NotExistent('The aiida user {} is not configured to use computer {}'.format(
                user.email, computer.name))
        except MultipleResultsFound:
            raise exceptions.ConfigurationError('The aiida user {} is configured more than once to use '
                                                'computer {}! Only one configuration is allowed'.format(
                                                    user.email, computer.name))

    def delete(self, authinfo_id):
        session = get_scoped_session()
        try:
            session.query(DbAuthInfo).filter_by(id=authinfo_id).one().delete()
            session.commit()
        except NoResultFound:
            raise exceptions.NotExistent("AuthInfo with id '{}' not found".format(authinfo_id))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the calcfunction decorator and CalcFunctionNode."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.backends.testbase import AiidaTestCase
from aiida.common import exceptions
from aiida.common.links import LinkType
from aiida.engine import calcfunction, Process
from aiida.manage.caching import enable_caching
from aiida.orm import Int, CalcFunctionNode


class TestCalcFunction(AiidaTestCase):
    """Tests for calcfunctions.

    .. note: tests common to all process functions should go in `aiida.backends.tests.work.test_process_function.py`
    """

    def setUp(self):
        super(TestCalcFunction, self).setUp()
        self.assertIsNone(Process.current())
        self.default_int = Int(256)

        @calcfunction
        def test_calcfunction(data):
            return Int(data.value + 1)

        self.test_calcfunction = test_calcfunction

    def tearDown(self):
        super(TestCalcFunction, self).tearDown()
        self.assertIsNone(Process.current())

    def test_calcfunction_node_type(self):
        """Verify that a calcfunction gets a CalcFunctionNode as node instance."""
        _, node = self.test_calcfunction.run_get_node(self.default_int)
        self.assertIsInstance(node, CalcFunctionNode)

    def test_calcfunction_links(self):
        """Verify that a calcfunction can only get CREATE links and no RETURN links."""
        _, node = self.test_calcfunction.run_get_node(self.default_int)

        self.assertEqual(len(node.get_outgoing(link_type=LinkType.CREATE).all()), 1)
        self.assertEqual(len(node.get_outgoing(link_type=LinkType.RETURN).all()), 0)

    def test_calcfunction_return_stored(self):
        """Verify that a calcfunction will raise when a stored node is returned."""

        @calcfunction
        def test_calcfunction():
            return Int(2).store()

        with self.assertRaises(ValueError):
            test_calcfunction.run_get_node()

    def test_calcfunction_default_linkname(self):
        """Verify that a calcfunction that returns a single Data node gets a default link label."""
        _, node = self.test_calcfunction.run_get_node(self.default_int)

        self.assertEqual(node.outputs.result, self.default_int.value + 1)
        self.assertEqual(getattr(node.outputs, Process.SINGLE_OUTPUT_LINKNAME), self.default_int.value + 1)
        self.assertEqual(node.outputs[Process.SINGLE_OUTPUT_LINKNAME], self.default_int.value + 1)

    def test_calcfunction_caching(self):
        """Verify that a calcfunction can be cached."""
        _, original = self.test_calcfunction.run_get_node(Int(5))

        # Caching a CalcFunctionNode should be possible
        with enable_caching(CalcFunctionNode):
            _, cached = self.test_calcfunction.run_get_node(Int(5))
            self.assertTrue(cached.is_created_from_cache)
            self.assertIn(cached.get_cache_source(), original.uuid)

    def test_calcfunction_caching_change_code(self):
        """Verify that changing the source codde of a calcfunction invalidates any existing cached nodes."""
        result_original = self.test_calcfunction(self.default_int)

        with enable_caching(CalcFunctionNode):

            @calcfunction
            def test_calcfunction(data):
                """This calcfunction has a different source code from the one setup in the setUp method."""
                return Int(data.value + 2)

            result_cached, cached = test_calcfunction.run_get_node(self.default_int)
            self.assertNotEqual(result_original, result_cached)
            self.assertFalse(cached.is_created_from_cache)

    def test_calcfunction_do_not_store_provenance(self):
        """Run the function without storing the provenance."""
        data = Int(1)
        result, node = self.test_calcfunction.run_get_node(data, metadata={'store_provenance': False})  # pylint: disable=unexpected-keyword-arg
        self.assertFalse(result.is_stored)
        self.assertFalse(data.is_stored)
        self.assertFalse(node.is_stored)
        self.assertEqual(result, data + 1)

    def test_calculation_cannot_call(self):
        """Verify that calling another process from within a calcfunction raises as it is forbidden."""

        @calcfunction
        def test_calcfunction_caller(data):
            self.test_calcfunction(data)

        with self.assertRaises(exceptions.InvalidOperation):
            test_calcfunction_caller(self.default_int)

    def test_calculation_call_store_provenance_false(self):  # pylint: disable=invalid-name
        """Verify that a `calcfunction` can call another calcfunction as long as `store_provenance` is False."""

        @calcfunction
        def test_calcfunction_caller(data):
            return self.test_calcfunction(data, metadata={'store_provenance': False})  # pylint: disable=unexpected-keyword-arg

        result, node = test_calcfunction_caller.run_get_node(self.default_int)

        self.assertTrue(isinstance(result, Int))
        self.assertTrue(isinstance(node, CalcFunctionNode))

        # The node of the outermost `calcfunction` should have a single `CREATE` link and no `CALL_CALC` links
        self.assertEqual(len(node.get_outgoing(link_type=LinkType.CREATE).all()), 1)
        self.assertEqual(len(node.get_outgoing(link_type=LinkType.CALL_CALC).all()), 0)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Sqla query builder implementation"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from datetime import datetime
import uuid
import six

# pylint: disable=no-name-in-module, import-error
from sqlalchemy_utils.types.choice import Choice
from sqlalchemy import and_, or_, not_
from sqlalchemy.types import Integer, Float, Boolean, DateTime
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.sql.expression import case, FunctionElement
from sqlalchemy.ext.compiler import compiles

import aiida.backends.sqlalchemy
from aiida.common.exceptions import InputValidationError
from aiida.orm.implementation.querybuilder import BackendQueryBuilder


class jsonb_array_length(FunctionElement):  # pylint: disable=invalid-name
    # pylint: disable=too-few-public-methods
    name = 'jsonb_array_len'


@compiles(jsonb_array_length)
def compile(element, compiler, **_kw):  # pylint: disable=function-redefined, redefined-builtin
    """
    Get length of array defined in a JSONB column
    """
    return "jsonb_array_length(%s)" % compiler.process(element.clauses)


class array_length(FunctionElement):  # pylint: disable=invalid-name
    # pylint: disable=too-few-public-methods
    name = 'array_len'


@compiles(array_length)
def compile(element, compiler, **_kw):  # pylint: disable=function-redefined
    """
    Get length of array defined in a JSONB column
    """
    return "array_length(%s)" % compiler.process(element.clauses)


class jsonb_typeof(FunctionElement):  # pylint: disable=invalid-name
    # pylint: disable=too-few-public-methods
    name = 'jsonb_typeof'


@compiles(jsonb_typeof)
def compile(element, compiler, **_kw):  # pylint: disable=function-redefined
    """
    Get length of array defined in a JSONB column
    """
    return "jsonb_typeof(%s)" % compiler.process(element.clauses)


class SqlaQueryBuilder(BackendQueryBuilder):
    """
    QueryBuilder to use with SQLAlchemy-backend and
    schema defined in backends.sqlalchemy.models
    """

    # pylint: disable=redefined-outer-name, too-many-public-methods

    @property
    def Node(self):
        import aiida.backends.sqlalchemy.models.node
        return aiida.backends.sqlalchemy.models.node.DbNode

    @property
    def Link(self):
        import aiida.backends.sqlalchemy.models.node
        return aiida.backends.sqlalchemy.models.node.DbLink

    @property
    def Computer(self):
        import aiida.backends.sqlalchemy.models.computer
        return aiida.backends.sqlalchemy.models.computer.DbComputer

    @property
    def User(self):
        import aiida.backends.sqlalchemy.models.user
        return aiida.backends.sqlalchemy.models.user.DbUser

    @property
    def Group(self):
        import aiida.backends.sqlalchemy.models.group
        return aiida.backends.sqlalchemy.models.group.DbGroup

    @property
    def AuthInfo(self):
        import aiida.backends.sqlalchemy.models.authinfo
        return aiida.backends.sqlalchemy.models.authinfo.DbAuthInfo

    @property
    def Comment(self):
        import aiida.backends.sqlalchemy.models.comment
        return aiida.backends.sqlalchemy.models.comment.DbComment

    @property
    def Log(self):
        import aiida.backends.sqlalchemy.models.log
        return aiida.backends.sqlalchemy.models.log.DbLog

    @property
    def table_groups_nodes(self):
        import aiida.backends.sqlalchemy.models.group
        return aiida.backends.sqlalchemy.models.group.table_groups_nodes

    def get_session(self):
        return aiida.backends.sqlalchemy.get_scoped_session()

    def modify_expansions(self, alias, expansions):
        """
        For sqlalchemy, there are no additional expansions for now, so
        I am returning an empty list
        """
        # pylint: disable=protected-access
        if issubclass(alias._sa_class_manager.class_, self.Computer) or \
            issubclass(alias._sa_class_manager.class_, self.Log):
            try:
                expansions.remove('metadata')
                expansions.append('_metadata')
            except KeyError:
                pass

        return expansions

    def get_filter_expr(self, operator, value, attr_key, is_attribute, alias=None, column=None, column_name=None):
        """
        Applies a filter on the alias given.
        Expects the alias of the ORM-class on which to filter, and filter_spec.
        Filter_spec contains the specification on the filter.
        Expects:

        :param operator: The operator to apply, see below for further details
        :param value:
            The value for the right side of the expression,
            the value you want to compare with.

        :param path: The path leading to the value

        :param attr_key: Boolean, whether the value is in a json-column,
            or in an attribute like table.


        Implemented and valid operators:

        *   for any type:
            *   ==  (compare single value, eg: '==':5.0)
            *   in    (compare whether in list, eg: 'in':[5, 6, 34]
        *  for floats and integers:
            *   >
            *   <
            *   <=
            *   >=
        *  for strings:
            *   like  (case - sensitive), for example
                'like':'node.calc.%'  will match node.calc.relax and
                node.calc.RELAX and node.calc. but
                not node.CALC.relax
            *   ilike (case - unsensitive)
                will also match node.CaLc.relax in the above example

            .. note::
                The character % is a reserved special character in SQL,
                and acts as a wildcard. If you specifically
                want to capture a ``%`` in the string, use: ``_%``

        *   for arrays and dictionaries (only for the
            SQLAlchemy implementation):

            *   contains: pass a list with all the items that
                the array should contain, or that should be among
                the keys, eg: 'contains': ['N', 'H'])
            *   has_key: pass an element that the list has to contain
                or that has to be a key, eg: 'has_key':'N')

        *  for arrays only (SQLAlchemy version):
            *   of_length
            *   longer
            *   shorter

        All the above filters invoke a negation of the
        expression if preceded by **~**::

            # first example:
            filter_spec = {
                'name' : {
                    '~in':[
                        'halle',
                        'lujah'
                    ]
                } # Name not 'halle' or 'lujah'
            }

            # second example:
            filter_spec =  {
                'id' : {
                    '~==': 2
                }
            } # id is not 2
        """
        # pylint: disable=too-many-arguments, too-many-branches
        expr = None
        if operator.startswith('~'):
            negation = True
            operator = operator.lstrip('~')
        elif operator.startswith('!'):
            negation = True
            operator = operator.lstrip('!')
        else:
            negation = False
        if operator in ('longer', 'shorter', 'of_length'):
            if not isinstance(value, int):
                raise InputValidationError("You have to give an integer when comparing to a length")
        elif operator in ('like', 'ilike'):
            if not isinstance(value, six.string_types):
                raise InputValidationError("Value for operator {} has to be a string (you gave {})"
                                           "".format(operator, value))

        elif operator == 'in':
            value_type_set = set(type(i) for i in value)
            if len(value_type_set) > 1:
                raise InputValidationError('{}  contains more than one type'.format(value))
            elif not value_type_set:
                raise InputValidationError('{}  contains is an empty list'.format(value))
        elif operator in ('and', 'or'):
            expressions_for_this_path = []
            for filter_operation_dict in value:
                for newoperator, newvalue in filter_operation_dict.items():
                    expressions_for_this_path.append(
                        self.get_filter_expr(
                            newoperator,
                            newvalue,
                            attr_key=attr_key,
                            is_attribute=is_attribute,
                            alias=alias,
                            column=column,
                            column_name=column_name))
            if operator == 'and':
                expr = and_(*expressions_for_this_path)
            elif operator == 'or':
                expr = or_(*expressions_for_this_path)

        if expr is None:
            if is_attribute:
                expr = self.get_filter_expr_from_attributes(
                    operator, value, attr_key, column=column, column_name=column_name, alias=alias)
            else:
                if column is None:
                    if (alias is None) and (column_name is None):
                        raise RuntimeError("I need to get the column but do not know the alias and the column name")
                    column = self.get_column(column_name, alias)
                expr = self.get_filter_expr_from_column(operator, value, column)

        if negation:
            return not_(expr)
        return expr

    def get_filter_expr_from_attributes(self, operator, value, attr_key, column=None, column_name=None, alias=None):
        # Too many everything!
        # pylint: disable=too-many-branches, too-many-arguments, too-many-statements

        def cast_according_to_type(path_in_json, value):
            """Cast the value according to the type"""
            if isinstance(value, bool):
                type_filter = jsonb_typeof(path_in_json) == 'boolean'
                casted_entity = path_in_json.astext.cast(Boolean)
            elif isinstance(value, (int, float)):
                type_filter = jsonb_typeof(path_in_json) == 'number'
                casted_entity = path_in_json.astext.cast(Float)
            elif isinstance(value, dict) or value is None:
                type_filter = jsonb_typeof(path_in_json) == 'object'
                casted_entity = path_in_json.astext.cast(JSONB)  # BOOLEANS?
            elif isinstance(value, dict):
                type_filter = jsonb_typeof(path_in_json) == 'array'
                casted_entity = path_in_json.astext.cast(JSONB)  # BOOLEANS?
            elif isinstance(value, six.string_types):
                type_filter = jsonb_typeof(path_in_json) == 'string'
                casted_entity = path_in_json.astext
            elif value is None:
                type_filter = jsonb_typeof(path_in_json) == 'null'
                casted_entity = path_in_json.astext.cast(JSONB)  # BOOLEANS?
            elif isinstance(value, datetime):
                # type filter here is filter whether this attributes stores
                # a string and a filter whether this string
                # is compatible with a datetime (using a regex)
                #  - What about historical values (BC, or before 1000AD)??
                #  - Different ways to represent the timezone

                type_filter = jsonb_typeof(path_in_json) == 'string'
                regex_filter = path_in_json.astext.op("SIMILAR TO")(
                    "\d\d\d\d-[0-1]\d-[0-3]\dT[0-2]\d:[0-5]\d:\d\d\.\d+((\+|\-)\d\d:\d\d)?")  # pylint: disable=anomalous-backslash-in-string
                type_filter = and_(type_filter, regex_filter)
                casted_entity = path_in_json.cast(DateTime)
            else:
                raise TypeError('Unknown type {}'.format(type(value)))
            return type_filter, casted_entity

        if column is None:
            column = self.get_column(column_name, alias)

        database_entity = column[tuple(attr_key)]
        if operator == '==':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity == value)], else_=False)
        elif operator == '>':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity > value)], else_=False)
        elif operator == '<':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity < value)], else_=False)
        elif operator in ('>=', '=>'):
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity >= value)], else_=False)
        elif operator in ('<=', '=<'):
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity <= value)], else_=False)
        elif operator == 'of_type':
            # http://www.postgresql.org/docs/9.5/static/functions-json.html
            #  Possible types are object, array, string, number, boolean, and null.
            valid_types = ('object', 'array', 'string', 'number', 'boolean', 'null')
            if value not in valid_types:
                raise InputValidationError("value {} for of_type is not among valid types\n"
                                           "{}".format(value, valid_types))
            expr = jsonb_typeof(database_entity) == value
        elif operator == 'like':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity.like(value))], else_=False)
        elif operator == 'ilike':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity.ilike(value))], else_=False)
        elif operator == 'in':
            type_filter, casted_entity = cast_according_to_type(database_entity, value[0])
            expr = case([(type_filter, casted_entity.in_(value))], else_=False)
        elif operator == 'contains':
            expr = database_entity.cast(JSONB).contains(value)
        elif operator == 'has_key':
            expr = database_entity.cast(JSONB).has_key(value)  # noqa
        elif operator == 'of_length':
            expr = case(
                [(jsonb_typeof(database_entity) == 'array', jsonb_array_length(database_entity.cast(JSONB)) == value)],
                else_=False)

        elif operator == 'longer':
            expr = case(
                [(jsonb_typeof(database_entity) == 'array', jsonb_array_length(database_entity.cast(JSONB)) > value)],
                else_=False)
        elif operator == 'shorter':
            expr = case(
                [(jsonb_typeof(database_entity) == 'array', jsonb_array_length(database_entity.cast(JSONB)) < value)],
                else_=False)
        else:
            raise InputValidationError("Unknown operator {} for filters in JSON field".format(operator))
        return expr

    def get_projectable_attribute(self, alias, column_name, attrpath, cast=None, **kwargs):
        """
        :returns: An attribute store in a JSON field of the give column
        """
        entity = self.get_column(column_name, alias)[attrpath]
        if cast is None:
            entity = entity
        elif cast == 'f':
            entity = entity.astext.cast(Float)
        elif cast == 'i':
            entity = entity.astext.cast(Integer)
        elif cast == 'b':
            entity = entity.astext.cast(Boolean)
        elif cast == 't':
            entity = entity.astext
        elif cast == 'j':
            entity = entity.astext.cast(JSONB)
        elif cast == 'd':
            entity = entity.astext.cast(DateTime)
        else:
            raise InputValidationError("Unkown casting key {}".format(cast))
        return entity

    def get_aiida_res(self, key, res):
        """
        Some instance returned by ORM (django or SA) need to be converted
        to Aiida instances (eg nodes). Choice (sqlalchemy_utils)
        will return their value

        :param key: The key
        :param res: the result returned by the query

        :returns: an aiida-compatible instance
        """
        if isinstance(res, Choice):
            returnval = res.value
        elif isinstance(res, uuid.UUID):
            returnval = six.text_type(res)
        else:
            try:
                returnval = self._backend.get_backend_entity(res)
            except TypeError:
                returnval = res

        return returnval

    def yield_per(self, query, batch_size):
        """
        :param count: Number of rows to yield per step

        Yields *count* rows at a time

        :returns: a generator
        """
        try:
            return query.yield_per(batch_size)
        except Exception:
            # exception was raised. Rollback the session
            self.get_session().rollback()
            raise

    def count(self, query):
        try:
            return query.count()
        except Exception:
            # exception was raised. Rollback the session
            self.get_session().rollback()
            raise

    def first(self, query):
        """
        Executes query in the backend asking for one instance.

        :returns: One row of aiida results
        """
        try:
            return query.first()
        except Exception:
            # exception was raised. Rollback the session
            self.get_session().rollback()
            raise

    def iterall(self, query, batch_size, tag_to_index_dict):
        if not tag_to_index_dict:
            raise Exception("Got an empty dictionary: {}".format(tag_to_index_dict))

        try:
            results = query.yield_per(batch_size)

            if len(tag_to_index_dict) == 1:
                # Sqlalchemy, for some strange reason, does not return a list of lsits
                # if you have provided an ormclass

                if list(tag_to_index_dict.values()) == ['*']:
                    for rowitem in results:
                        yield [self.get_aiida_res(tag_to_index_dict[0], rowitem)]
                else:
                    for rowitem, in results:
                        yield [self.get_aiida_res(tag_to_index_dict[0], rowitem)]
            elif len(tag_to_index_dict) > 1:
                for resultrow in results:
                    yield [
                        self.get_aiida_res(tag_to_index_dict[colindex], rowitem)
                        for colindex, rowitem in enumerate(resultrow)
                    ]
            else:
                raise ValueError("Got an empty dictionary")
        except Exception:
            self.get_session().rollback()
            raise

    def iterdict(self, query, batch_size, tag_to_projected_entity_dict):

        nr_items = sum(len(v) for v in tag_to_projected_entity_dict.values())

        if not nr_items:
            raise ValueError("Got an empty dictionary")

        # Wrapping everything in an atomic transaction:
        try:
            results = query.yield_per(batch_size)
            if nr_items > 1:
                for this_result in results:
                    yield {
                        tag: {
                            attrkey: self.get_aiida_res(attrkey, this_result[index_in_sql_result])
                            for attrkey, index_in_sql_result in projected_entities_dict.items()
                        } for tag, projected_entities_dict in tag_to_projected_entity_dict.items()
                    }
            elif nr_items == 1:
                # I this case, sql returns a  list, where each listitem is the result
                # for one row. Here I am converting it to a list of lists (of length 1)
                if [v for entityd in tag_to_projected_entity_dict.values() for v in entityd.keys()] == ['*']:
                    for this_result in results:
                        yield {
                            tag: {
                                attrkey: self.get_aiida_res(attrkey, this_result)
                                for attrkey, position in projected_entities_dict.items()
                            } for tag, projected_entities_dict in tag_to_projected_entity_dict.items()
                        }
                else:
                    for this_result, in results:
                        yield {
                            tag: {
                                attrkey: self.get_aiida_res(attrkey, this_result)
                                for attrkey, position in projected_entities_dict.items()
                            } for tag, projected_entities_dict in tag_to_projected_entity_dict.items()
                        }
            else:
                raise ValueError("Got an empty dictionary")
        except Exception:
            self.get_session().rollback()
            raise



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
"""Utilities for the workflow engine."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import contextlib
import logging

from six.moves import range
import tornado.ioloop
from tornado import concurrent, gen

__all__ = ('RefObjectStore', 'interruptable_task', 'InterruptableFuture')

LOGGER = logging.getLogger(__name__)
PROCESS_STATE_CHANGE_KEY = 'process|state_change|{}'
PROCESS_STATE_CHANGE_DESCRIPTION = 'The last time a process of type {}, changed state'


class InterruptableFuture(concurrent.Future):
    """A future that can be interrupted by calling `interrupt`."""

    def interrupt(self, reason):
        """This method should be called to interrupt the coroutine represented by this InterruptableFuture."""
        self.set_exception(reason)

    @gen.coroutine
    def with_interrupt(self, yieldable):
        """
        Yield a yieldable which will be interrupted if this future is interrupted ::

            from tornado import ioloop, gen
            loop = ioloop.IOLoop.current()

            interruptable = InterutableFuture()
            loop.add_callback(interruptable.interrupt, RuntimeError("STOP"))
            loop.run_sync(lambda: interruptable.with_interrupt(gen.sleep(2)))
            >>> RuntimeError: STOP


        :param yieldable: The yieldable
        :return: The result of the yieldable
        """
        # Wait for one of the two to finish, if it's us that finishes we expect that it was
        # because of an exception that will have been raised automatically
        wait_iterator = gen.WaitIterator(yieldable, self)
        result = yield wait_iterator.next()  # pylint: disable=stop-iteration-return
        if not wait_iterator.current_index == 0:
            raise RuntimeError("This interruptible future had it's result set unexpectedly to {}".format(result))

        result = yield [yieldable, self][0]
        raise gen.Return(result)


def interruptable_task(coro, loop=None):
    """
    Turn the given coroutine into an interruptable task by turning it into an InterruptableFuture and returning it.

    :param coro: the coroutine that should be made interruptable
    :param loop: the event loop in which to run the coroutine, by default uses tornado.ioloop.IOLoop.current()
    :return: an InterruptableFuture
    """

    loop = loop or tornado.ioloop.IOLoop.current()
    future = InterruptableFuture()

    @gen.coroutine
    def execute_coroutine():
        """Coroutine that wraps the original coroutine and sets it result on the future only if not already set."""
        try:
            result = yield coro(future)
        except Exception as exception:  # pylint: disable=broad-except
            if not future.done():
                future.set_exception(exception)
        else:
            # If the future has not been set elsewhere, i.e. by the interrupt call, by the time that the coroutine
            # is executed, set the future's result to the result of the coroutine
            if not future.done():
                future.set_result(result)

    loop.add_callback(execute_coroutine)

    return future


def ensure_coroutine(fct):
    """
    Ensure that the given function ``fct`` is a coroutine

    If the passed function is not already a coroutine, it will be made to be a coroutine

    :param fct: the function
    :returns: the coroutine
    """
    if tornado.gen.is_coroutine_function(fct):
        return fct

    @tornado.gen.coroutine
    def wrapper(*args, **kwargs):
        raise tornado.gen.Return(fct(*args, **kwargs))

    return wrapper


@gen.coroutine
def exponential_backoff_retry(fct, initial_interval=10.0, max_attempts=5, logger=None, ignore_exceptions=None):
    """
    Coroutine to call a function, recalling it with an exponential backoff in the case of an exception

    This coroutine will loop ``max_attempts`` times, calling the ``fct`` function, breaking immediately when the call
    finished without raising an exception, at which point the returned result will be raised, wrapped in a
    ``tornado.gen.Result`` instance. If an exception is caught, the function will yield a ``tornado.gen.sleep`` with a
    time interval equal to the ``initial_interval`` multiplied by ``2*N`` where ``N`` is the number of excepted calls.

    :param fct: the function to call, which will be turned into a coroutine first if it is not already
    :param initial_interval: the time to wait after the first caught exception before calling the coroutine again
    :param max_attempts: the maximum number of times to call the coroutine before re-raising the exception
    :raises: ``tornado.gen.Result`` if the ``coro`` call completes within ``max_attempts`` retries without raising
    """
    if logger is None:
        logger = LOGGER

    result = None
    coro = ensure_coroutine(fct)
    interval = initial_interval

    for iteration in range(max_attempts):
        try:
            result = yield coro()
            break  # Finished successfully
        except Exception as exception:  # pylint: disable=broad-except

            # Re-raise exceptions that should be ignored
            if ignore_exceptions is not None and isinstance(exception, ignore_exceptions):
                raise

            if iteration == max_attempts - 1:
                logger.warning('maximum attempts %d of calling %s, exceeded', max_attempts, coro.__name__)
                raise
            else:
                logger.exception('iteration %d of %s excepted, retrying after %d seconds', iteration + 1, coro.__name__,
                                 interval)

                yield gen.sleep(interval)
                interval *= 2

    raise gen.Return(result)


def is_process_function(function):
    """
    Return whether the given function is a process function

    :param function: a function
    :returns: True if the function is a wrapped process function, False otherwise
    """
    try:
        return function.is_process_function
    except AttributeError:
        return False


@contextlib.contextmanager
def loop_scope(loop):
    """
    Make an event loop current for the scope of the context

    :param loop: The event loop to make current for the duration of the scope
    :type loop: :class:`tornado.ioloop.IOLoop`
    """
    current = tornado.ioloop.IOLoop.current()

    try:
        loop.make_current()
        yield
    finally:
        current.make_current()


def set_process_state_change_timestamp(process):
    """
    Set the global setting that reflects the last time a process changed state, for the process type
    of the given process, to the current timestamp. The process type will be determined based on
    the class of the calculation node it has as its database container.

    :param process: the Process instance that changed its state
    """
    from aiida.backends.utils import set_global_setting
    from aiida.common.exceptions import UniquenessError
    from aiida.orm import ProcessNode, CalculationNode, WorkflowNode
    from aiida.common import timezone

    if isinstance(process.node, CalculationNode):
        process_type = 'calculation'
    elif isinstance(process.node, WorkflowNode):
        process_type = 'work'
    elif isinstance(process.node, ProcessNode):
        # This will only occur for testing, as in general users cannot launch plain Process classes
        return
    else:
        raise ValueError('unsupported calculation node type {}'.format(type(process.node)))

    key = PROCESS_STATE_CHANGE_KEY.format(process_type)
    description = PROCESS_STATE_CHANGE_DESCRIPTION.format(process_type)
    value = timezone.now()

    try:
        set_global_setting(key, value, description)
    except UniquenessError as exception:
        process.logger.debug('could not update the {} setting because of a UniquenessError: {}'.format(key, exception))


def get_process_state_change_timestamp(process_type=None):
    """
    Get the global setting that reflects the last time a process of the given process type changed its state.
    The returned value will be the corresponding timestamp or None if the setting does not exist.

    :param process_type: optional process type for which to get the latest state change timestamp.
        Valid process types are either 'calculation' or 'work'. If not specified, last timestamp for all
        known process types will be returned.
    :return: a timestamp or None
    """
    from aiida.backends.utils import get_global_setting

    valid_process_types = ['calculation', 'work']

    if process_type is not None and process_type not in valid_process_types:
        raise ValueError("invalid value for process_type, valid values are {}".format(', '.join(valid_process_types)))

    if process_type is None:
        process_types = valid_process_types
    else:
        process_types = [process_type]

    timestamps = []

    for process_type_key in process_types:
        key = PROCESS_STATE_CHANGE_KEY.format(process_type_key)
        try:
            timestamps.append(get_global_setting(key))
        except KeyError:
            pass

    if not timestamps:
        return None

    return max(timestamps)


class RefObjectStore(object):  # pylint: disable=useless-object-inheritance
    """
    An object store that has a reference count based on a context manager.
    Basic usage::

        store = RefObjectStore()
        with store.get('Martin', lambda: 'martin.uhrin@epfl.ch') as email:
            with store.get('Martin') as email2:
                email is email2  # True

    The use case for this store is when you have an object can be used by
    multiple parts of the code simultaneously (nested or async code) and
    where there should be one instance that exists for the lifetime of these
    contexts.  Once noone is using the object, it should be removed from the
    store (and therefore eventually garbage collected).
    """

    class Reference(object):  # pylint: disable=useless-object-inheritance
        """A reference to store the context reference count and the object itself."""

        def __init__(self, obj):
            self._count = 0
            self._obj = obj

        @property
        def count(self):
            """
            Get the reference count for the object

            :return: The reference count
            :rtype: int
            """
            return self._count

        @contextlib.contextmanager
        def get(self):
            """
            Get the object itself, which will up the reference count for the duration of the context.

            :return: The object
            """
            self._count += 1
            try:
                yield self._obj
            finally:
                self._count -= 1

    def __init__(self):
        self._objects = {}

    @contextlib.contextmanager
    def get(self, identifier, constructor=None):
        """
        Get or create an object.  The internal reference count will be upped for
        the duration of the context.  If the reference count drops to 0 the object
        will be automatically removed from the list.

        :param identifier: The key identifying the object
        :param constructor: An optional constructor that is called with no arguments
            if the object doesn't already exist in the store
        :return: The object corresponding to the identifier
        """
        try:
            ref = self._objects[identifier]
        except KeyError:
            if constructor is None:
                raise ValueError("Object not found and no constructor given")
            ref = self.Reference(constructor())
            self._objects[identifier] = ref

        try:
            with ref.get() as obj:
                yield obj
        finally:
            if ref.count == 0:
                self._objects.pop(identifier)



#!/usr/bin/env runaiida
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This example illustrates in a very minimal way how a WorkChain can be defined and how it can be run. This mostly
illustrates how the spec of the WorkChain is defined and how functions in the outline of the spec have to be defined.
"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from aiida.engine import WorkChain, run
from aiida.orm import NumericType, Float, Int


class SumWorkChain(WorkChain):

    @classmethod
    def define(cls, spec):
        super(SumWorkChain, cls).define(spec)
        spec.input('a', valid_type=NumericType)
        spec.input('b', valid_type=NumericType)
        spec.outline(
            cls.sum
        )
        spec.output('sum', valid_type=NumericType)

    def sum(self):
        self.out('sum', self.inputs.a + self.inputs.b)


class ProductWorkChain(WorkChain):

    @classmethod
    def define(cls, spec):
        super(ProductWorkChain, cls).define(spec)
        spec.input('a', valid_type=NumericType)
        spec.input('b', valid_type=NumericType)
        spec.outline(
            cls.product
        )
        spec.output('product', valid_type=NumericType)

    def product(self):
        self.out('product', self.inputs.a * self.inputs.b)


class SumProductWorkChain(WorkChain):

    @classmethod
    def define(cls, spec):
        super(SumProductWorkChain, cls).define(spec)
        spec.input('a', valid_type=NumericType)
        spec.input('b', valid_type=NumericType)
        spec.input('c', valid_type=NumericType)
        spec.outline(
            cls.sum,
            cls.product
        )
        spec.output('sumproduct', valid_type=NumericType)

    def sum(self):
        self.ctx.sum = self.inputs.a + self.inputs.b

    def product(self):
        self.out('sumproduct', self.ctx.sum * self.inputs.c)


def main():
    inputs = {
        'a': Float(3.14),
        'b': Int(4),
        'c': Int(6)
    }

    results = run(SumWorkChain, **inputs)
    print('Result of SumWorkChain: {}'.format(results))

    results = run(ProductWorkChain, **inputs)
    print('Result of ProductWorkChain: {}'.format(results))

    results = run(SumProductWorkChain, **inputs)
    print('Result of SumProductWorkChain: {}'.format(results))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Module for the non empty string parameter type
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from click.types import StringParamType


class NonEmptyStringParamType(StringParamType):
    """
    Parameter that cannot be an an empty string.
    """
    name = 'nonemptystring'

    def convert(self, value, param, ctx):
        newval = super(NonEmptyStringParamType, self).convert(value, param, ctx)
        if not newval:  # None or empty string
            self.fail("Empty string is not valid!")

        return newval

    def __repr__(self):
        return 'NONEMPTYSTRING'



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import datetime

import plumpy
from tornado import gen

from aiida.backends.testbase import AiidaTestCase
from aiida.backends.tests.utils import processes as test_processes
from aiida.engine import ProcessState, submit
from aiida.manage.manager import get_manager
from aiida.orm import Int


class TestProcessControl(AiidaTestCase):
    """
    Test AiiDA's RabbitMQ functionalities.
    """

    TIMEOUT = 2.

    def setUp(self):
        super(TestProcessControl, self).setUp()

        # These two need to share a common event loop otherwise the first will never send
        # the message while the daemon is running listening to intercept
        manager = get_manager()
        self.runner = manager.get_runner()
        self.daemon_runner = manager.create_daemon_runner(loop=self.runner.loop)

    def tearDown(self):
        self.daemon_runner.close()
        super(TestProcessControl, self).tearDown()

    def test_submit_simple(self):
        # Launch the process
        @gen.coroutine
        def do_submit():
            calc_node = submit(test_processes.DummyProcess)
            yield self.wait_for_calc(calc_node)

            self.assertTrue(calc_node.is_finished_ok)
            self.assertEqual(calc_node.process_state.value, plumpy.ProcessState.FINISHED.value)

        self.runner.loop.run_sync(do_submit)

    def test_launch_with_inputs(self):

        @gen.coroutine
        def do_launch():
            a = Int(5)
            b = Int(10)

            calc_node = submit(test_processes.AddProcess, a=a, b=b)
            yield self.wait_for_calc(calc_node)
            self.assertTrue(calc_node.is_finished_ok)
            self.assertEqual(calc_node.process_state.value, plumpy.ProcessState.FINISHED.value)

        self.runner.loop.run_sync(do_launch)

    def test_submit_bad_input(self):
        with self.assertRaises(ValueError):
            submit(test_processes.AddProcess, a=Int(5))

    def test_exception_process(self):

        @gen.coroutine
        def do_exception():
            calc_node = submit(test_processes.ExceptionProcess)
            yield self.wait_for_calc(calc_node)

            self.assertFalse(calc_node.is_finished_ok)
            self.assertEqual(calc_node.process_state.value, plumpy.ProcessState.EXCEPTED.value)

        self.runner.loop.run_sync(do_exception)

    def test_pause(self):
        """Testing sending a pause message to the process."""

        controller = get_manager().get_process_controller()

        @gen.coroutine
        def do_pause():
            calc_node = submit(test_processes.WaitProcess)
            while calc_node.process_state != ProcessState.WAITING:
                yield

            self.assertFalse(calc_node.paused)

            future = yield with_timeout(controller.pause_process(calc_node.pk))
            result = yield self.wait_future(future)
            self.assertTrue(result)
            self.assertTrue(calc_node.paused)

        self.runner.loop.run_sync(do_pause)

    def test_pause_play(self):
        """Test sending a pause and then a play message."""

        controller = get_manager().get_process_controller()

        @gen.coroutine
        def do_pause_play():
            calc_node = submit(test_processes.WaitProcess)
            self.assertFalse(calc_node.paused)
            while calc_node.process_state != ProcessState.WAITING:
                yield

            pause_message = 'Take a seat'
            future = yield with_timeout(controller.pause_process(calc_node.pk, msg=pause_message))
            result = yield self.wait_future(future)
            self.assertTrue(calc_node.paused)
            self.assertEqual(calc_node.process_status, pause_message)

            future = yield with_timeout(controller.play_process(calc_node.pk))
            result = yield self.wait_future(future)
            self.assertTrue(result)
            self.assertFalse(calc_node.paused)
            self.assertEqual(calc_node.process_status, None)

        self.runner.loop.run_sync(do_pause_play)

    def test_kill(self):
        """Test sending a kill message."""

        controller = get_manager().get_process_controller()

        @gen.coroutine
        def do_kill():
            calc_node = submit(test_processes.WaitProcess)
            self.assertFalse(calc_node.is_killed)
            while calc_node.process_state != ProcessState.WAITING:
                yield

            kill_message = 'Sorry, you have to go mate'
            future = yield with_timeout(controller.kill_process(calc_node.pk, msg=kill_message))
            result = yield self.wait_future(future)
            self.assertTrue(result)

            self.wait_for_calc(calc_node)
            self.assertTrue(calc_node.is_killed)
            self.assertEqual(calc_node.process_status, kill_message)

        self.runner.loop.run_sync(do_kill)

    @gen.coroutine
    def wait_for_calc(self, calc_node, timeout=2.):
        future = self.runner.get_calculation_future(calc_node.pk)
        raise gen.Return((yield with_timeout(future, timeout)))

    @gen.coroutine
    def wait_future(self, future, timeout=2.):
        raise gen.Return((yield with_timeout(future, timeout)))


@gen.coroutine
def with_timeout(what, timeout=5.0):
    raise gen.Return((yield gen.with_timeout(datetime.timedelta(seconds=timeout), what)))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Base class for CalculationTools

Sub-classes can be registered in the `aiida.tools.calculations` category to enable the `CalcJobNode` class from being
able to find the tools plugin, load it and expose it through the `tools` property of the `CalcJobNode`.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

__all__ = ('CalculationTools',)


class CalculationTools(object):
    """Base class for CalculationTools."""

    # pylint: disable=too-few-public-methods,useless-object-inheritance

    def __init__(self, node):
        self._node = node



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for the extended dictionary classes."""
# pylint: disable=pointless-statement,attribute-defined-outside-init
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import copy
import pickle
import unittest

from six import integer_types

from aiida.common import json
from aiida.common import exceptions
from aiida.common import extendeddicts


class TestFFADExample(extendeddicts.FixedFieldsAttributeDict):
    """
    An example class that accepts only the 'alpha', 'beta' and 'gamma' keys/attributes.
    """
    _valid_fields = ('alpha', 'beta', 'gamma')


class TestDFADExample(extendeddicts.DefaultFieldsAttributeDict):
    """
    An example class that has 'alpha', 'beta' and 'gamma' as default keys.
    """
    _default_fields = ('alpha', 'beta', 'gamma')

    @staticmethod
    def validate_alpha(value):
        """Validate a value."""
        # Ok if unset
        if value is None:
            return

        if not isinstance(value, integer_types):
            raise TypeError('expecting integer')
        if value < 0:
            raise ValueError('expecting a positive or zero value')


class TestAttributeDictAccess(unittest.TestCase):
    """
    Try to access the dictionary elements in various ways, copying (shallow and
    deep), check raised exceptions.
    """

    def test_access_dict_to_attr(self):
        """Test dictionary to attribute."""
        dictionary = extendeddicts.AttributeDict()
        dictionary['test'] = 'abc'
        self.assertEqual(dictionary.test, 'abc')

    def test_access_attr_to_dict(self):
        """Test attribute to dictionary."""
        dictionary = extendeddicts.AttributeDict()
        dictionary.test = 'def'
        self.assertEqual(dictionary['test'], 'def')

    def test_access_nonexisting_asattr(self):
        """Test non-existing attribute."""
        dictionary = extendeddicts.AttributeDict()
        with self.assertRaises(AttributeError):
            dictionary.test

    def test_access_nonexisting_askey(self):
        """Test non-existing attribute as key."""
        dictionary = extendeddicts.AttributeDict()
        with self.assertRaises(KeyError):
            dictionary['test']

    def test_del_nonexisting_askey(self):
        """Test deleting non-existing attribute as key."""
        dictionary = extendeddicts.AttributeDict()
        with self.assertRaises(KeyError):
            del dictionary['test']

    def test_del_nonexisting_asattr(self):
        """Test deleting non-existing attribute."""
        dictionary = extendeddicts.AttributeDict()
        with self.assertRaises(AttributeError):
            del dictionary.test

    def test_copy(self):
        """Test copying."""
        dictionary_01 = extendeddicts.AttributeDict()
        dictionary_01.alpha = 'a'
        dictionary_02 = dictionary_01.copy()
        dictionary_02.alpha = 'b'
        self.assertEqual(dictionary_01.alpha, 'a')
        self.assertEqual(dictionary_02.alpha, 'b')

    def test_delete_after_copy(self):
        """Test deleting after copying."""
        dictionary_01 = extendeddicts.AttributeDict()
        dictionary_01.alpha = 'a'
        dictionary_01.beta = 'b'
        dictionary_02 = dictionary_01.copy()
        del dictionary_01.alpha
        del dictionary_01['beta']
        with self.assertRaises(AttributeError):
            _ = dictionary_01.alpha
        with self.assertRaises(KeyError):
            _ = dictionary_01['beta']
        self.assertEqual(dictionary_02['alpha'], 'a')
        self.assertEqual(dictionary_02.beta, 'b')
        self.assertEqual(set(dictionary_01.keys()), set({}))
        self.assertEqual(set(dictionary_02.keys()), set({'alpha', 'beta'}))

    def test_shallowcopy1(self):
        """Test shallow copying."""
        dictionary_01 = extendeddicts.AttributeDict()
        dictionary_01.alpha = [1, 2, 3]
        dictionary_01.beta = 3
        dictionary_02 = dictionary_01.copy()
        dictionary_02.alpha[0] = 4
        dictionary_02.beta = 5
        self.assertEqual(dictionary_01.alpha, [4, 2, 3])  # copy does a shallow copy
        self.assertEqual(dictionary_02.alpha, [4, 2, 3])
        self.assertEqual(dictionary_01.beta, 3)
        self.assertEqual(dictionary_02.beta, 5)

    def test_shallowcopy2(self):
        """Test shallow copying."""
        dictionary_01 = extendeddicts.AttributeDict()
        dictionary_01.alpha = {'a': 'b', 'c': 'd'}
        # dictionary_02 = copy.deepcopy(dictionary_01)
        dictionary_02 = dictionary_01.copy()
        # doesn't work like this, would work as dictionary_02['x']['a']
        # i think that it is because deepcopy on dict actually creates a
        # copy only if the data is changed; but for a nested dict,
        # dictionary_02.alpha returns a dict wrapped in our class and this looses all the
        # information on what should be updated when changed.
        dictionary_02.alpha['a'] = 'ggg'
        self.assertEqual(dictionary_01.alpha['a'], 'ggg')  # copy does a shallow copy
        self.assertEqual(dictionary_02.alpha['a'], 'ggg')

    def test_deepcopy1(self):
        """Test deep copying."""
        dictionary_01 = extendeddicts.AttributeDict()
        dictionary_01.alpha = [1, 2, 3]
        dictionary_02 = copy.deepcopy(dictionary_01)
        dictionary_02.alpha[0] = 4
        self.assertEqual(dictionary_01.alpha, [1, 2, 3])
        self.assertEqual(dictionary_02.alpha, [4, 2, 3])

    def test_shallowcopy3(self):
        """Test shallow copying."""
        dictionary_01 = extendeddicts.AttributeDict()
        dictionary_01.alpha = {'a': 'b', 'c': 'd'}
        dictionary_02 = copy.deepcopy(dictionary_01)
        dictionary_02.alpha['a'] = 'ggg'
        self.assertEqual(dictionary_01.alpha['a'], 'b')  # copy does a shallow copy
        self.assertEqual(dictionary_02.alpha['a'], 'ggg')


class TestAttributeDictNested(unittest.TestCase):
    """
    Test the functionality of nested AttributeDict classes.
    """

    def test_nested(self):
        """Test nested dictionary."""
        dictionary_01 = extendeddicts.AttributeDict({'x': 1, 'y': 2})
        dictionary_02 = extendeddicts.AttributeDict({'z': 3, 'w': 4})
        dictionary_01.nested = dictionary_02
        self.assertEqual(dictionary_01.nested.z, 3)
        self.assertEqual(dictionary_01['nested'].w, 4)
        self.assertEqual(dictionary_01.nested['w'], 4)
        dictionary_02['w'] = 5
        self.assertEqual(dictionary_01['nested'].w, 5)
        self.assertEqual(dictionary_01.nested['w'], 5)

    def test_comparison(self):
        """Test dictionary comparison."""
        dictionary_01 = extendeddicts.AttributeDict({'x': 1, 'y': 2, 'z': extendeddicts.AttributeDict({'w': 3})})
        dictionary_02 = extendeddicts.AttributeDict({'x': 1, 'y': 2, 'z': extendeddicts.AttributeDict({'w': 3})})

        # They compare to the same value but they are different objects
        self.assertFalse(dictionary_01 is dictionary_02)
        self.assertEqual(dictionary_01, dictionary_02)

        dictionary_02.z.w = 4
        self.assertNotEqual(dictionary_01, dictionary_02)

    def test_nested_deepcopy(self):
        """Test nested deepcopy."""
        dictionary_01 = extendeddicts.AttributeDict({'x': 1, 'y': 2})
        dictionary_02 = extendeddicts.AttributeDict({'z': 3, 'w': 4})
        dictionary_01.nested = dictionary_02
        dictionary_01copy = copy.deepcopy(dictionary_01)
        self.assertEqual(dictionary_01copy.nested.z, 3)
        self.assertEqual(dictionary_01copy['nested'].w, 4)
        self.assertEqual(dictionary_01copy.nested['w'], 4)
        dictionary_02['w'] = 5
        self.assertEqual(dictionary_01copy['nested'].w, 4)  # Nothing has changed
        self.assertEqual(dictionary_01copy.nested['w'], 4)  # Nothing has changed
        self.assertEqual(dictionary_01copy.nested.w, 4)  # Nothing has changed

        self.assertEqual(dictionary_01['nested'].w, 5)  # The old one is updated
        self.assertEqual(dictionary_01.nested['w'], 5)  # The old one is updated
        self.assertEqual(dictionary_01.nested.w, 5)  # The old one is updated

        dictionary_01copy.nested.w = 6
        self.assertEqual(dictionary_01copy.nested.w, 6)
        self.assertEqual(dictionary_01.nested.w, 5)
        self.assertEqual(dictionary_02.w, 5)


class TestAttributeDictSerialize(unittest.TestCase):
    """
    Test serialization/deserialization (with json, pickle, ...)
    """

    def test_json(self):
        """Test loading and dumping from json."""
        dictionary_01 = extendeddicts.AttributeDict({'x': 1, 'y': 2})
        dictionary_02 = json.loads(json.dumps(dictionary_01))
        # Note that here I am comparing a dictionary (dictionary_02) with a
        # extendeddicts.AttributeDict (dictionary_02) and they still compare to equal
        self.assertEqual(dictionary_01, dictionary_02)

    def test_json_recursive(self):
        """Test loading and dumping from json recursively."""
        dictionary_01 = extendeddicts.AttributeDict({'x': 1, 'y': 2, 'z': extendeddicts.AttributeDict({'w': 4})})
        dictionary_02 = json.loads(json.dumps(dictionary_01))
        # Note that here I am comparing a dictionary (dictionary_02) with a (recursive)
        # extendeddicts.AttributeDict (dictionary_02) and they still compare to equal
        self.assertEqual(dictionary_01, dictionary_02)

    def test_pickle(self):
        """Test pickling."""
        dictionary_01 = extendeddicts.AttributeDict({'x': 1, 'y': 2})
        dictionary_02 = pickle.loads(pickle.dumps(dictionary_01))
        self.assertEqual(dictionary_01, dictionary_02)

    def test_pickle_recursive(self):
        """Test pickling recursively."""
        dictionary_01 = extendeddicts.AttributeDict({'x': 1, 'y': 2, 'z': extendeddicts.AttributeDict({'w': 4})})
        dictionary_02 = pickle.loads(pickle.dumps(dictionary_01))
        self.assertEqual(dictionary_01, dictionary_02)


class TestFFAD(unittest.TestCase):
    """Test for the fixed fields attribute dictionary."""

    def test_insertion(self):
        """Test insertion."""
        dictionary = TestFFADExample()
        dictionary['alpha'] = 1
        dictionary.beta = 2
        # Not a valid key.
        with self.assertRaises(KeyError):
            dictionary['delta'] = 2
        with self.assertRaises(AttributeError):
            dictionary.epsilon = 5

    def test_insertion_on_init(self):
        """Test insertion in constructor."""
        TestFFADExample({'alpha': 1, 'beta': 2})
        with self.assertRaises(KeyError):
            # 'delta' is not a valid key
            TestFFADExample({'alpha': 1, 'delta': 2})

    def test_pickle(self):
        """Note: pickle works here because self._valid_fields is defined at the class level!"""
        dictionary_01 = TestFFADExample({'alpha': 1, 'beta': 2})
        dictionary_02 = pickle.loads(pickle.dumps(dictionary_01))
        dictionary_02.gamma = 3
        with self.assertRaises(KeyError):
            dictionary_02['delta'] = 2

    def test_class_attribute(self):
        """
        I test that the get_valid_fields() is working as a class method,
        so I don't need to instantiate the class to get the list.
        """
        self.assertEqual(set(TestFFADExample.get_valid_fields()), set(['alpha', 'beta', 'gamma']))


class TestDFAD(unittest.TestCase):
    """Test for the default fields attribute dictionary."""

    def test_insertion_and_retrieval(self):
        """Test insertion and retrieval."""
        dictionary = TestDFADExample()
        dictionary['alpha'] = 1
        dictionary.beta = 2
        dictionary['delta'] = 3
        dictionary.epsilon = 4
        self.assertEqual(dictionary.alpha, 1)
        self.assertEqual(dictionary.beta, 2)
        self.assertEqual(dictionary['delta'], 3)
        self.assertEqual(dictionary['epsilon'], 4)

    def test_keylist_method(self):
        """Test keylist retrieval."""
        dictionary = TestDFADExample()
        dictionary['alpha'] = 1
        dictionary.beta = 2
        dictionary['delta'] = 3
        dictionary.epsilon = 4

        self.assertEqual(set(dictionary.get_default_fields()), set(['alpha', 'beta', 'gamma']))
        self.assertEqual(set(dictionary.keys()), set(['alpha', 'beta', 'delta', 'epsilon']))
        self.assertEqual(set(dictionary.defaultkeys()), set(['alpha', 'beta']))
        self.assertEqual(set(dictionary.extrakeys()), set(['delta', 'epsilon']))
        self.assertIsNone(dictionary.gamma)

    def test_class_attribute(self):
        """
        I test that the get_default_fields() is working as a class method,
        so I don't need to instantiate the class to get the list.
        """
        self.assertEqual(set(TestDFADExample.get_default_fields()), set(['alpha', 'beta', 'gamma']))

    def test_validation(self):
        """Test validation."""
        dictionary = TestDFADExample()

        # Should be ok to have an empty 'alpha' attribute
        dictionary.validate()

        dictionary.alpha = 4
        dictionary.beta = 'text'

        # This should be fine
        dictionary.validate()

        dictionary.alpha = 'string'
        # dictionary.alpha must be a positive integer
        with self.assertRaises(exceptions.ValidationError):
            dictionary.validate()

        dictionary.alpha = -3
        # alpha must be a positive integer
        with self.assertRaises(exceptions.ValidationError):
            dictionary.validate()
